"""Analog System and Digital Control Simulator

This module provides simulator tools to simulate the hardware
interaction between an analog system and digital control.
These are mainly intended to produce control signals
:math:`\mathbf{s}[k]` and evaluate state vector trajectories
:math:`\mathbf{x}(t)` for various Analog system
:py:class:`cbadc.analog_system.AnalogSystem` and
:py:class:`cbadc.digital_control.DigitalControl` interactions.
"""
from .analog_system import AnalogSystem
from .digital_control import DigitalControl
from .analog_signal import AnalogSignal, ConstantSignal, Sinusodial
import numpy as np
from scipy.integrate import solve_ivp
from scipy.linalg import expm
import math
from typing import Iterator, Generator, List, Dict, Union


class StateSpaceSimulator(Iterator[np.ndarray]):
    """Simulate the analog system and digital control interactions
    in the presence on analog signals.

    Parameters
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        the analog system
    digital_control: :py:class:`cbadc.digital_control.DigitalControl`
        the digital control
    input_signals : [:py:class:`cbadc.analog_signal.AnalogSignal`]
        a python list of analog signals (or a derived class)
    Ts : `float`, optional
        specify a sampling rate at which we want to evaluate the systems
        , defaults to :py:class:`digitalControl.Ts`. Note that this Ts must be smaller
        than :py:class:`digitalControl.Ts`.
    t_stop : `float`, optional
        determines a stop time, defaults to :py:obj:`math.inf`



    Attributes
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        the analog system being simulated.
    digital_control : :py:class:`cbadc.digital_control.DigitalControl`
        the digital control being simulated.
    t : `float`
        current time of simulator.
    Ts : `float`
        sample rate of simulation.
    t_stop : `float`
        end time at which the generator raises :py:class:`StopIteration`.
    rtol, atol : `float`, `optional`
        Relative and absolute tolerances. The solver keeps the local error estimates less
        than atol + rtol * abs(y). Effects the underlying solver as described in
        :py:func:`scipy.integrate.solve_ivp`. Default to 1e-3 for rtol and 1e-6 for atol.
    max_step : `float`, `optional`
        Maximum allowed step size. Default is np.inf, i.e., the step size is not
        bounded and determined solely by the solver. Effects the underlying solver as
        described in :py:func:`scipy.integrate.solve_ivp`. Defaults to :py:obj:`math.inf`.
    initial_state_vector: `array_like`, shape=(N), `optional`
        initial state vector.
    See also
    --------
    :py:class:`cbadc.analog_signal.AnalogSignal`
    :py:class:`cbadc.analog_system.AnalogSystem`
    :py:class:`cbadc.digital_control.DigitalControl`

    Examples
    --------
    >>> from cbadc.simulator import StateSpaceSimulator
    >>> from cbadc.analog_signal import Sinusodial
    >>> from cbadc.analog_system import AnalogSystem
    >>> from cbadc.digital_control import DigitalControl
    >>> import numpy as np
    >>> A = np.array([[0., 0], [6250., 0.]])
    >>> B = np.array([[6250., 0]]).transpose()
    >>> CT = np.array([[1, 0], [0, 1]])
    >>> Gamma = np.array([[-6250, 0], [0, -6250]])
    >>> Gamma_tildeT = CT
    >>> analog_system = AnalogSystem(A, B, CT, Gamma, Gamma_tildeT)
    >>> digital_control = DigitalControl(1e-6, 2)
    >>> input_signal = Sinusodial(1.0, 250)
    >>> simulator = StateSpaceSimulator(analog_system, digital_control, (input_signal,))
    >>> _ = simulator.__next__()
    >>> _ = simulator.__next__()
    >>> print(np.array(simulator.__next__()))
    [0 0]

    See also
    --------

    Yields
    ------
    `array_like`, shape=(M,), dtype=numpy.int8

    Raises
    ------
    str : unknown

    """

    def __init__(self,
                 analog_system: AnalogSystem,
                 digital_control: DigitalControl,
                 input_signal: List[Union[AnalogSignal, ConstantSignal,
                                          Sinusodial]],
                 Ts: float = None,
                 t_stop: float = math.inf,
                 atol: float = 1e-12,
                 rtol: float = 1e-12,
                 max_step: float = math.inf,
                 initial_state_vector=None
                 ):
        if analog_system.L != len(input_signal):
            raise BaseException("""The analog system does not have as many inputs as in input
            list""")
        if not np.allclose(analog_system.D, np.zeros_like(analog_system.D)):
            raise BaseException(
                """Can't simulate system with non-zero
                D matrix. Consider chaining systems to remove D.""")
        self.analog_system = analog_system
        self.digital_control = digital_control
        self.input_signals = input_signal
        self.t: float = 0.
        self.t_stop = t_stop
        if Ts:
            self.Ts = Ts
        else:
            self.Ts = self.digital_control.T
        if self.Ts > self.digital_control.T:
            raise BaseException(
                f"Simulating with a sample period {self.Ts} that exceeds the control period of the digital control {self.digital_control.T}")
        if initial_state_vector is not None:
            self._state_vector = np.array(
                initial_state_vector, dtype=np.float64)
            print("initial state vector: ", self._state_vector)
            if self._state_vector.size != self.analog_system.N or len(self._state_vector.shape) > 1:
                raise BaseException(
                    "initial_state_vector not single dimension of length N")
        else:
            self._state_vector = np.zeros(
                self.analog_system.N, dtype=np.double)
        self._temp_state_vector = np.zeros(
            self.analog_system.N, dtype=np.double)
        self._control_observation = np.zeros(
            self.analog_system.M_tilde, dtype=np.double)
        self._input_vector = np.zeros(self.analog_system.L, dtype=np.double)
        self._control_vector = np.zeros(self.analog_system.M, dtype=np.double)
        self._res = np.zeros(self.analog_system.N, dtype=np.double)
        self.atol = atol  # 1e-6
        self.rtol = rtol  # 1e-6
        if (max_step > self.Ts):
            self.max_step = self.Ts / 1e-1
        else:
            self.max_step = max_step
        self._pre_computations()
        # self.solve_oder = self._ordinary_differential_solution
        # self.solve_oder = self._full_ordinary_differential_solution

    def state_vector(self) -> np.ndarray:
        """return current analog system state vector :math:`\mathbf{x}(t)`
        evaluated at time :math:`t`.

        Examples
        --------
        >>> from cbadc.simulator import StateSpaceSimulator
        >>> from cbadc.analog_signal import Sinusodial
        >>> from cbadc.analog_system import AnalogSystem
        >>> from cbadc.digital_control import DigitalControl
        >>> import numpy as np
        >>> A = np.array([[0., 0], [6250., 0.]])
        >>> B = np.array([[6250., 0]]).transpose()
        >>> CT = np.array([[1, 0], [0, 1]])
        >>> Gamma = np.array([[-6250, 0], [0, -6250]])
        >>> Gamma_tildeT = CT
        >>> analog_system = AnalogSystem(A, B, CT, Gamma, Gamma_tildeT)
        >>> digital_control = DigitalControl(1e-6, 2)
        >>> input_signal = Sinusodial(1.0, 250)
        >>> simulator = StateSpaceSimulator(analog_system, digital_control, (input_signal,))
        >>> _ = simulator.__next__()
        >>> _ = simulator.__next__()
        >>> print(np.array(simulator.state_vector()))
        [-0.00623036 -0.00626945]

        Returns
        -------
        `array_like`, shape=(N,)
            returns the state vector :math:`\mathbf{x}(t)`
        """
        return self._state_vector[:]

    def __iter__(self):
        """Use simulator as an iterator
        """
        return self

    def __next__(self) -> np.ndarray:
        """Computes the next control signal :math:`\mathbf{s}[k]`
        """
        t_end: float = self.t + self.Ts
        t_span = np.array((self.t, t_end))
        if t_end >= self.t_stop:
            raise StopIteration
        self._state_vector = self._ordinary_differential_solution(t_span)
        # self._state_vector = self._full_ordinary_differential_solution(t_span)
        self.t = t_end
        return self.digital_control.control_signal()

    def _analog_system_matrix_exponential(self, t: float) -> np.ndarray:
        return np.asarray(expm(np.asarray(self.analog_system.A) * t))

    def _pre_computations(self):
        """Precomputes quantities for quick evaluation of state transition and control
        contribution.

        Specifically,

        :math:`\exp\\left(\mathbf{A} T_s \\right)`

        and

        :math:`\mathbf{A}_c = \int_{0}^{T_s} \exp\\left(\mathbf{A} (T_s - \tau)\\right) \mathbf{\Gamma} \mathbf{d}(\tau) \mathrm{d} \tau`

        are computed where the formed describes the state transition and the latter
        the control contributions. Furthermore, :math:`\mathbf{d}(\tau)` is the DAC waveform
        (or impulse response) of the digital control.
        """

        # expm(A T_s)
        self._pre_computed_state_transition_matrix = self._analog_system_matrix_exponential(
            self.Ts)

        def derivative(t, x):
            dac_waveform = np.zeros(
                (self.analog_system.M, self.analog_system.M))
            for m in range(self.analog_system.M):
                dac_waveform[:, m] = self.digital_control.impulse_response(
                    m, t)
            return np.dot(
                np.asarray(
                    self._analog_system_matrix_exponential(self.Ts - t)),
                np.dot(np.asarray(self.analog_system.Gamma),
                       dac_waveform)
            ).flatten()

        tspan = np.array([0, self.Ts])
        atol = 1e-12
        rtol = 1e-12
        max_step = self.Ts * 1e-3

        self._pre_computed_control_matrix = solve_ivp(derivative,
                                                      tspan,
                                                      np.zeros(
                                                          (self.analog_system.N * self.analog_system.M)),
                                                      atol=atol, rtol=rtol, max_step=max_step
                                                      ).y[:, -1].reshape((self.analog_system.N, self.analog_system.M), order='C')

    def _ordinary_differential_solution(self, t_span: np.ndarray) -> np.ndarray:
        """Computes system ivp in three parts:

        First solve input signal contribution by computing

        :math:`\mathbf{u}_{c} = \int_{t_1}^{t_2} \mathbf{A} x(t) + \mathbf{B} \mathbf{u}(t) \mathrm{d} t`

        where :math:`\mathbf{x}(t_1) = \begin{pmatrix} 0, & \dots, & 0 \end{pmatrix}^{\mathsf{T}}`.

        Secondly advance the previous state as

        :math:`\mathbf{x}_c = \mathbf{x}(t_2) = \exp\\left( \mathbf{A} T_s \\right) \mathbf{x}(t_1)`

        Thirdly, compute the control contribution by

        :math:`\mathbf{s}_c = \mathbf{A}_c \mathbf{s}[k]`

        where :math:`\mathbf{A}_c = \int_{0}^{T_s} \exp\\left(\mathbf{A} (T_s - \tau)\\right) \mathbf{\Gamma} \mathbf{d}(\tau) \mathrm{d} \tau`
        and :math:`\mathbf{d}(\tau)` is the DAC waveform (or impulse response) of the digital control.

        Finally, all contributions are added and returned as

        :math:`\mathbf{u}_{c} + \mathbf{x}_c + \mathbf{s}_c`.

        Parameters
        ----------
        t_span : (float, float)
            the initial time :math:`t_1` and end time :math:`t_2` of the simulation.

        Returns
        -------
        array_like, shape=(N,)
            computed state vector.
        """

        # Compute signal contribution
        def f(t, x):
            return self._signal_derivative(t, x)
        self._temp_state_vector = solve_ivp(
            f,
            t_span,
            np.zeros(self.analog_system.N),
            atol=self.atol,
            rtol=self.rtol,
            max_step=self.max_step
        ).y[:, -1]

        self._temp_state_vector += np.dot(
            self._pre_computed_state_transition_matrix, self._state_vector).flatten()

        # Compute control observation s_tilde(t)
        self._control_observation = np.dot(
            self.analog_system.Gamma_tildeT, self._state_vector)

        # update control at time t_span[0]
        self._char_control_vector = self.digital_control.control_contribution(
            t_span[0], self._control_observation)
        self._temp_state_vector += np.dot(
            self._pre_computed_control_matrix, self._char_control_vector).flatten()
        return self._temp_state_vector

    def _signal_derivative(self, t: float, x: np.ndarray) -> np.ndarray:
        res = np.dot(self.analog_system.A, x)
        for _l in range(self.analog_system.L):
            res += np.dot(self.analog_system.B[:, _l],
                          self.input_signals[_l].evaluate(t))
        return res.flatten()

    def _full_ordinary_differential_solution(self, t_span: np.ndarray) -> np.ndarray:
        return solve_ivp(
            lambda t, x: self._ordinary_differentail_function(t, x),
            t_span,
            self._state_vector,
            atol=self.atol,
            rtol=self.rtol,
            max_step=self.max_step
        ).y[:, -1]

    def _ordinary_differentail_function(self, t: float, y: np.ndarray):
        """Solve the differential computational problem
        of the analog system and digital control interaction

        Parameters
        ----------
        t : `float`
            the time for evaluation
        y : array_lik, shape=(N,)
            state vector

        Returns
        -------
        array_like, shape=(N,)
            vector of derivatives evaluated at time t.
        """
        for _l in range(self.analog_system.L):
            self._input_vector[_l] = self.input_signals[_l].evaluate(t)
        self._temp_state_vector = np.dot(self.analog_system.Gamma_tildeT, y)
        self._control_vector = self.digital_control.control_contribution(
            t, self._temp_state_vector)
        return np.asarray(self.analog_system.derivative(self._temp_state_vector, t, self._input_vector, self._control_vector)).flatten()

    def __str__(self):
        return f"t = {self.t}, (current simulator time)\nTs = {self.Ts},\nt_stop = {self.t_stop},\nrtol = {self.rtol},\natol = {self.atol}, and\nmax_step = {self.max_step}\n"


def extended_simulation_result(simulator: StateSpaceSimulator) -> Generator[Dict[str, np.ndarray], None, None]:
    """Extended simulation output

    Used to also pass the state vector from a
    simulator generator.

    Parameters
    ----------
    simulator : :py:class:`cbadc.simulator.StateSpaceSimulator`
        a iterable simulator instance.

    Yields
    ------
    { 'control_signal', 'analog_state' } : { (array_like, shape=(M,)), (array_like, shape=(N,)) }
        an extended output including the analog state vector.
    """
    for control_signal in simulator:
        analog_state = simulator.state_vector()
        yield {
            'control_signal': np.array(control_signal),
            'analog_state': np.array(analog_state)
        }



"""
This submodule provides several software/hardware datasets.

These examples provide quick access to simulated control signal sequences from
hardware prototypes. Additionally, as simulating a control-bounded ADC is
typically a computationally intense task, compared to filtering, several control
signal sequences from default architectures and parametrizations are provided to
reduce simulation time.
"""

# from .hadamard import HadamardPCB
# from .chain_of_integrators import ChainOfIntegrators



"""digital estimator calibration.

this module contains tools for calibrating the digital estimator.
"""
from cbadc.digital_estimator import DigitalEstimator


class Calibrator:

    def __init__(self, digitalEstimator):
        self.digitalEstimator = digitalEstimator

    def cost_function(self):
        pass


class InputReference:

    def __init__(self, ):
        a = DigitalEstimator




"""Data wrappers for control-bounded Hadamard ADC datasets.
"""
import numpy as np
import cbadc
import scipy.linalg


class HadamardPCB():
    """The Hadamard PCB...

    Parameters
    ----------
    pcb: `str`
        string determining prototype version, currently
        pcb = {'A', 'B'}, defaults to 'B'.
    """

    def __init__(self, pcb="B"):
        """Initialize a HadamardPCB Simulation for a given PCB version
        """
        self.r_1 = 1.62e3
        self.r_2 = 3.24e3
        self.r_3 = 10e3
        self.r_4 = np.inf

        self.v_ctrl = 2.5
        self.pcb = pcb
        if pcb == "B":
            self.T = 1e-6
            self.c_1 = 3.3e-9
        elif pcb == "A":
            self.T = 1.0/24e6
            self.c_1 = 120e-12
        else:
            raise BaseException("Specified PCB version does not supported.")
        # Hadamard matrix
        H_n = scipy.linalg.hadamard(4)
        buffer_loss = self.r_3 / (4*self.r_3 + self.r_1)
        A_chain_of_integrators = np.diag(
            np.ones(3), k=-1) - self.r_1 / self.r_4 * np.eye(4, k=0)
        A = 1/(self.r_1 * self.c_1) * np.matmul(-H_n,
                                                np.matmul(A_chain_of_integrators,
                                                          -H_n * buffer_loss))
        B = -1/(self.r_1 * self.c_1) * np.ones((4, 1))
        CT = np.eye(4)
        # assuming inputs +/-1 from the digital control
        Gamma = -self.v_ctrl/(self.r_2 * self.c_1) * \
            np.hstack((-H_n, np.eye(4)))
        # scaling does not matter, sign is positive since inverting comparators
        # and inverting flipflip cancel out
        Gamma_tildeT = np.vstack((-H_n, np.eye(4)))
        self.analog_system = cbadc.analog_system.AnalogSystem(
            A, B, CT, Gamma, Gamma_tildeT)
        self.digital_control = cbadc.digital_control.DigitalControl(self.T, 8)

    def simulation_ramp_1_B(self):
        """20 mHz period length, 4.5 Vpp ramp input simulation

        Prototype:

        - Version B
        - R1 = 1.62 kOhm
        - R2 = 3.24 kOhm
        - R3 = 10 kOhm
        - R5 = np.inf
        - T = 1 / (1 MHz)
        - C = 3.3 nF

        Input signal:

        - type: ramp signal
        - period: 1/20 mHz
        - amplitude: 4.5 Vpp

        Simulation Setup:

        - signal generator: HP...
        - signal length: 100s (100M control signal samples)
        - total size: 100 MB

        Returns
        -------
        Iterator[np.ndarray, None, None]
            the measured control signal sequence from hardware
        Iterator[np.ndarray, None, None]
            an ideal control signal sequence from simulation
        :py:class:`cbadc.simulator.StateSpaceSimulator`
            an instantiated ideal simulator.

        """
        if self.pcb != "B":
            raise BaseException("This simulation was made with PCB-B")
        input_signal = cbadc.analog_signal.Ramp(4.5, 1/20e-3)
        size = int(100e6)
        ideal_simulator = cbadc.simulator.StateSpaceSimulator(
            self.analog_system, self.digital_control, [input_signal],
            t_stop=self.digital_control.T * size)
        control_signal_hardware = cbadc.utilities.byte_stream_2_control_signal(
            cbadc.utilities.read_byte_stream_from_url(
                'https://people.ee.ethz.ch/~merik/ramp_20mHz_4.5Vpp_1MHz_100s_pcbB.adc', self.analog_system.M),
            self.analog_system.M)
        control_signal_software_simulated = cbadc.utilities.byte_stream_2_control_signal(
            cbadc.utilities.read_byte_stream_from_url(
                'https://people.ee.ethz.ch/~merik/ideal_ramp_20mHz_4.5Vpp_1MHz_100s_pcbB.adc', self.analog_system.M),
            self.analog_system.M)
        return control_signal_hardware, control_signal_software_simulated, ideal_simulator, size



"""Utility functions

This module contains various helpful functions to accommodate
the cbadc toolbox.
"""
import struct
from typing import Generator, Iterator, Union
import numpy as np
from scipy.signal import welch
from typing import Tuple
from tqdm import tqdm
import requests
import os
import pickle
import scipy.io.wavfile
import numpy.typing as npt


def number_of_bytes_selector(M: int):
    """A helper function for selecting
    the right bytes settings.

    Parameters
    ----------
    M : int
        number of controls to be turned into bytes.

    Returns
    -------
    { number_of_bytes, format_marker, c_type_name} : {int, str, str}
        the byte type for conversion.
    """
    if M < 9:
        return {
            'number_of_bytes': 1,
            'format_marker': 'B',
            'c_type_name': 'unsigned char'
        }
    if M < 17:
        return {
            'number_of_bytes': 2,
            'format_marker': 'h',
            'c_type_name': 'short'
        }
    if M < 65:
        return {
            'number_of_bytes': 4,
            'format_marker': 'i',
            'c_type_name': 'int'
        }
    if M < 129:
        return {
            'number_of_bytes': 8,
            'format_marker': 'q',
            'c_type_name': 'long long'
        }
    raise BaseException(
        f"M={M} is larger than 128 which is the largest allowed size")


def control_signal_2_byte_stream(control_signal: Iterator[np.ndarray], M: int) -> Generator[bytes, None, None]:
    """Convert a control signal into a byte stream

    Parameters
    ----------
    control_signal : [array_like, shape=(M,)]
        a iterator producing sequences of control signals (binary).
    M : int
        number of control inputs.

    Yields
    ------
    bytes
        a binary string representing the control signal.

    """
    format = number_of_bytes_selector(M)
    format_marker = format['format_marker']
    for s in control_signal:
        sum = 0
        for m in range(M):
            # Construct integer by bit shifting
            if s[m] > 0:
                sum += 1 << m
        yield struct.pack(format_marker, sum)


def byte_stream_2_control_signal(byte_stream: Iterator[bytes], M: int) -> Generator[np.ndarray, None, None]:
    """Convert a byte stream into a control_sequence

    Parameters
    ----------
    byte_stream : binary buffer
        a byte stream iterator
    M : int
        number of control inputs.

    Yields
    -------
    array_like, shape=(M,)
        a control signal sample.

    Example
    -------
    >>> M = 3
    >>> control_signal = np.array([[0, 1, 0], [1, 0, 1],[0, 0, 1]])
    >>> cs =  byte_stream_2_control_signal(control_signal_2_byte_stream(control_signal, M), M)
    >>> next(cs)
    array([0, 1, 0], dtype=int8)
    >>> next(cs)
    array([1, 0, 1], dtype=int8)
    >>> next(cs)
    array([0, 0, 1], dtype=int8)
    """
    format = number_of_bytes_selector(M)
    format_marker = format['format_marker']
    mask = 1
    for bs in byte_stream:
        if (not bs):
            raise StopIteration
        sum = struct.unpack(format_marker, bs)[0]
        s = np.zeros(M, dtype=np.int8)
        for m in range(M):
            # populate s from sum
            s[m] = mask & (sum >> m)
        yield s


def write_byte_stream_to_file(filename: str, iterator: Iterator[bytes]):
    """Write a stream into binary file.

    Parameters
    ----------
    filename : `str`
        filename for output file
    iterator : [bytes]
        an iterator of bytes to write to file
    """
    with open(filename, "wb") as f:
        for word in iterator:
            f.write(word)


def write_byte_stream_to_files(filename: str, iterator: Iterator[bytes], words_per_file: int = 100000):
    """Write a stream into a sequence of binary files of size words_per_file.

    Parameters
    ----------
    filename : `str`
        filename for output file
    iterator : [bytes]
        an iterator of bytes to write to files
    words_per_file: `int`
        number of words to be written per file, defaults to 100000.
        For byte-sized words the default corresponds to 100MB files.
    """
    count = 0
    iteration = 0
    base, ext = os.path.splitext(filename)
    while True:
        name = base + f"_{iteration}" + ext
        with open(name, "wb") as f:
            while (count < words_per_file):
                f.write(next(iterator))
                count += 1
        count = 0
        iteration += 1


def read_byte_stream_from_file(filenames: Union[str, list], M: int) -> Generator[bytes, None, None]:
    """Generate a byte stream iterator from file

    Parameters
    ----------
    filename : `str`
        filename for input file
    M : `int`
        number of controls

    Yields
    ------
    bytes :
        returns bytes
    """
    format = number_of_bytes_selector(M)
    if type(filenames) is str:
        filenames = [filenames]
    for filename in filenames:
        with open(filename, "rb") as f:
            byte = b'0'
            try:
                while byte:
                    byte = f.read(format['number_of_bytes'])
                    yield byte
            except StopIteration:
                pass
    raise StopIteration


def read_byte_stream_from_url(urlstring: Union[str, list], M: int) -> Generator[bytes, None, None]:
    """Generate a byte stream iterator from http request

    Parameters
    ----------
    urlstring : `str`
        url for input file
    M : `int`
        number of controls

    Yields
    ------
    bytes :
        returns bytes
    """
    session = requests.Session()
    format = number_of_bytes_selector(M)
    urls = []
    if type(urlstring) is str:
        urls = [urlstring]
    else:
        urls = urlstring
    for url in urls:
        try:
            request = session.get(url, stream=True)
            for chunk in request.iter_content(chunk_size=format['number_of_bytes'], decode_unicode=False):
                yield chunk
        except StopIteration:
            pass
    raise StopIteration


def random_control_signal(M: int, stop_after_number_of_iterations: int = (1 << 63), random_seed: int = 0) -> Generator[np.ndarray, None, None]:
    """Creates a iterator producing random control signals.

    Parameters
    ----------
    M : `int`
        number of controls
    stop_after_number_of_iterations : `int`, `optional`
        number of iterations until :py:class:`StopIteration` is raised,
        defaults to  :math:`2^{63}`.
    random_seed : `int`, `optional`
        used for setting the random seed, see :py:class:`numpy.random.seed`.

    Yields
    ------
    array_like, shape=(M,)
        a random control signal
    """
    if random_seed:
        np.random.seed(random_seed)
    iteration: int = 0
    while (iteration < stop_after_number_of_iterations):
        iteration += 1
        yield np.random.randint(2, size=M, dtype=np.int8)
    else:
        raise StopIteration


def compute_power_spectral_density(sequence: np.ndarray, nperseg: int = 1 << 14, fs: float = 1.0) -> Tuple[np.ndarray, np.ndarray]:
    """Compute power spectral density of sequence.

    Parameters
    ----------
    sequence : array_like, shape=(L, K)
        determine length of each fft sequence
    fs : `double`, `optional`
        sampling frequency, defaults to 1.0

    Returns
    -------
    ((array_like, shape=(K,)), (array_like, shape=(L, K)))
        frequencies [Hz] and PSD [:math:`V^2/\mathrm{Hz}`] of sequence.
    """
    freq, spectrum = welch(
        sequence,
        window='hanning',
        nperseg=nperseg,
        noverlap=None,
        nfft=None,
        return_onesided=True,
        scaling='density',
        fs=fs
    )
    return (np.asarray(freq), np.asarray(spectrum))


def snr_spectrum_computation(spectrum: np.ndarray,
                             signal_mask: np.ndarray,
                             noise_mask: np.ndarray):
    """Compute snr from spectrum

    Parameters
    ----------
    spectrum: ndarray
        a frequency spectrum
    signal_mask: ndarray
        an array containing the indices corresponding to the inband signal
        components.
    noise_mask: ndarray
        an array containing the indices corresponding to the inband noise.

    Returns
    -------
    float
        the signal-to-noise ratio.
    """
    noise = np.sum(spectrum[noise_mask])
    signal = np.sum(spectrum[signal_mask])
    if (noise > 0):
        return signal/noise
    else:
        return np.inf


def find_sinusoidal(spectrum: np.ndarray, mask_width: np.ndarray):
    """Find the peak in the spectrum and return indexes.

    Parameters
    ----------
    spectrum: ndarray
        a power spectral density or equivalent
    mask_width: `int`
        the width around peak to be considered.

    """
    candidate_peak = np.argmax(np.abs(spectrum))
    return np.arange(candidate_peak - mask_width // 2, candidate_peak + mask_width // 2)


def show_status(iterator, length: int = 1 << 63):
    """Write progress to stdout using :py:mod:`tqdm`.

    Parameters
    ----------
    iterator : a iterator
        a iterator to yield values from.
    length : int
        indicate length of iterator. Also causes a raises a StopIteration
        if iteration exceeds length, defaults to 2^63.
    """
    iterator_with_progress = tqdm(iterator)
    if length < (1 << 63):
        iterator_with_progress.total = length
    for iteration, value in enumerate(iterator_with_progress):
        if not iteration < length:
            raise StopIteration
        yield value


def pickle_dump(object_to_be_pickled, filename: str):
    """Convenience function to pickle an object.

    In principle all class instances in this package
    can be pickled to a file for later use.

    For example this includes:

    - analog signal instances,
    - analog system instances,
    - digital control instance,
    - simulation instances,
    - and digital estimator instances.

    See also
    --------
    :py:func:`cbadc.utilities.pickle_load`

    Parameters
    ----------
    object_to_be_pickled: any
        a pickleable object
    filename: `str`
        the path for it to be stored.
    """
    with open(filename, 'wb') as f:
        pickle.dump(object_to_be_pickled, f, protocol=-1)


def pickle_load(filename: str):
    """Convenience function to load python object from pickled file

    See also
    --------
    :py:func:`cbadc.utilities.pickle_dump`

    Parameters
    ----------
    filename: `str`
        the filename of the pickled file.
    """
    with open(filename, 'rb') as f:
        return pickle.load(f)


def iterator_to_numpy_array(iterator: Iterator[bytes], size: int, L: int = 1):
    """Convert an iterator into a numpy array

    Parameters
    ----------
    iterator: 
        a iterator with data points to fill up numpy array.
    size: `int`
        length of numpy array.
    L: `int`
        dimenson of the datapoints.

    Returns
    -------
    array_like, shape=(size, L)
    """
    if size < 1 or L < 1:
        raise BaseException("Both size and L must be positive integers.")
    data = np.zeros((size, L), dtype=np.double)
    for index in range(size):
        data[index, :] = next(iterator)
    return data


def write_wave(filename: str, sample_rate: int, data: npt.ArrayLike):
    """Create wave file from data array.

    This is a wrapper function for :py:func:`scipy.io.wavefile.write`.

    Parameters
    ----------
    filename: `str`
        name of the file to be generated
    sample_rate: `int`
        the sample rate in samples/second
    data: `array_like, shape=(number of samples, number of channels)
        the data array to be encoded as wave file
    """
    scipy.io.wavfile.write(filename, sample_rate, data)



"""
The control-bounded converter toolbox allows simulation and reconstruction
of control-bounded converters.
"""
from importlib.metadata import version
from . import analog_signal
from . import analog_system
from . import digital_control
from . import digital_estimator
from . import simulator
from . import utilities

# Set version variable
__version__ = version('cbadc')



"""Predefined common analog signals

This module focuses on representing analog signals, i.e., mappings from the
time :math:`t` to a signal value :math:`u(t)`. Typically for signal processing
algorithms, we are used to handeling discrete-time signals, i.e. samples of
signals. However, since the control-bounded A/D converters are converting
continuous-time signals we need tools to define signals that can be evaluated
over their whole continuous domain.
"""
import numpy as np


class AnalogSignal:
    """A default continuous-time analog signal.
    """

    def __init__(self):
        pass

    def evaluate(self, t: float) -> float:
        """Evaluate the signal at time :math:`t`.

        Parameters
        ----------
        t : `float`
            the time instance for evaluation.

        Returns
        -------
        float
            The analog signal value
        """
        return 0.0

    def __str__(self):
        return "Analog signal returns constant 0, i.e., maps t |-> 0."


class ConstantSignal(AnalogSignal):
    """A constant continuous-time analog signal.

    Parameters
    -----------
    offset : `float`, `optional`
         Determines the offset or DC bias of the analog signal, defaults to 0.


    Attributes
    ----------
    offset : `float`
        The offset value.


    See also
    ---------
    :py:class:`cbadc.analog_signal.Sinusodial`
    :py:class:`cbadc.simulator.StateSpaceSimulator`

    Examples
    ---------
    >>> from cbadc.analog_signal import ConstantSignal
    >>> u = ConstantSignal(1.0)
    >>> print(u.evaluate(42))
    1.0

    """

    def __init__(self, offset: float = 0.0):
        """Create a constant analog signal.
        """
        self.offset: float = offset

    def evaluate(self, t: float) -> float:
        """Evaluate the signal at time :math:`t`.

        Parameters
        ----------
        t : `float`
            the time instance for evaluation.

        Returns
        -------
        float
            The analog signal value
        """
        return self.offset

    def __str__(self):
        return f"ConstantSignal has an offset = {self.offset}."


class Sinusodial(AnalogSignal):
    """An analog continuous-time sinusodial signal.

    Parameters
    ----------
    amplitude : `float`
        The amplitude of the sinusodial.
    frequency : `float`
        Frequency of the oscillation in [Hz].
    phase : `float`, optional
        The phase, defaults to 0.
    offset : `float`
        The offset value.

    Attributes
    ----------
    amplitude : `float`
        The amplitude.
    frequency : `float`
        The frequency in [Hz].
    angularFrequency : `float`
        The frequency in [radians/second].
    phase : `float`
        The phase.
    offset : `float`, `optional`
        The offset

    See also
    --------
    cbadc.analog_signal.AnalogSignal

    Example
    -------
    >>> from cbadc.analog_signal import Sinusodial
    >>> import numpy as np
    >>> u = Sinusodial(1, 123, np.pi/2, 0)
    >>> print(u.evaluate(0))
    1.0

    """

    def __init__(self, amplitude: float, frequency: float, phase: float = 0.0,
                 offset: float = 0.0):
        self.amplitude: float = amplitude
        self.frequency: float = frequency
        self.angularFrequency: float = 2 * np.pi * self.frequency
        self.phase: float = phase
        self.offset: float = offset

    def __str__(self):
        return f"""Sinusodial parameterized as: \namplitude = {self.amplitude}, \n
        frequency = {self.frequency}, \nphase = {self.phase},
        and\noffset = {self.offset}"""

    def evaluate(self, t: float) -> float:
        """Evaluate the signal at time :math:`t`.

        Parameters
        ----------
        t : `float`
            the time instance for evaluation.

        Returns
        -------
        float
            The analog signal value
        """
        return (
            self.amplitude *
            np.sin(self.angularFrequency * t + self.phase) + self.offset
        )


class Ramp(AnalogSignal):
    """An analog continuous-time ramp signal.
    Parameters
    ----------
    amplitude : `float`
        The amplitude of the sinusodial.
    period : `float`
        one over the period length of the ramp, specified in [Hz].
    phase : `float`, optional
        The phase offset in [s], defaults to 0.
    offset : `float`
        The offset value.
    Attributes
    ----------
    amplitude : `float`
        The amplitude.
    period : `float`
        The ramp's time period in [s].
    phase : `float`
        The phase offset in [s].
    offset : `float`, `optional`
        The offset
    See also
    --------
    cbadc.analog_signal.AnalogSignal
    """

    def __init__(self, amplitude: float, period: float, phase: float = 0.0,
                 offset: float = 0.0):
        self.amplitude: float = amplitude
        self.period: float = period
        self.phase: float = phase
        self.offset: float = offset - self.amplitude / 2.0

    def __str__(self):
        return f"""Sinusodial parameterized as: \namplitude = {self.amplitude}, \n
        frequency = {self.frequency}, \nphase = {self.phase},
        and\noffset = {self.offset}"""

    def evaluate(self, t: float) -> float:
        """Evaluate the signal at time :math:`t`.
        Parameters
        ----------
        t : `float`
            the time instance for evaluation.
        Returns
        -------
        float
            The analog signal value
        """
        return(self.amplitude *
               ((t + self.phase) % self.period) + self.offset)


class SincPulse(AnalogSignal):

    """An analog continuous-time sinc pulse.

    Parameters
    ----------
    amplitude : `float`
        The peak amplitude of the pulse.
    bandwidth : `float`
        The bandwidth in [Hz].
    delay : `float`
        The time delay (instance of the peak) in [s].
    offset : `float`
        The offset value, defaults to 0.

    Attributes
    ----------
    amplitude : `float`
        The amplitude.
    bandwidth : `float`
        The bandwidth in [Hz].
    delay : `float`
        The the time delay in [s].
    offset : `float`, `optional`
        The offset

    See also
    --------
    cbadc.analog_signal.AnalogSignal

    Example
    -------
    >>> from cbadc.analog_signal import SincPulse
    >>> import numpy as np
    >>> u = SincPulse(3, 1, 5)
    >>> print(u.evaluate(5))
    3.0

    """

    def __init__(self, amplitude: float, bandwidth: float, delay: float,  
                 offset: float = 0.0):
        self.amplitude: float = amplitude
        self.bandwidth: float = bandwidth
        self.delay: float = delay
        self.offset: float = offset

    def __str__(self):
        return f"""Sinc pulse parameterized as: delay = {self.delay}, \n
        bandwidth = {self.bandwidth}, peak amplitude = {self.amplitude},  
        and\noffset = {self.offset}"""
    
    def evaluate(self, t: float) -> float:
        """Evaluate the signal at time :math:`t`.

        Parameters
        ----------
        t : `float`
            the time instance for evaluation.

        Returns
        -------
        float
            The analog signal value
        """
        return (
            self.amplitude * 
            np.sinc(2 * self.bandwidth * (t - self.delay)) +
            self.offset
        )




"""Chain of integrators datasets.
"""
import numpy as np
import cbadc
import logging
logger = logging.getLogger(__name__)

# map (N, beta, rho, kappa, input_signal_type, amplitude, frequency, phase, offset) -> url_string
chain_of_integrators_pre_simulations = {
    (1, 6250.0, 0.0, 1.0, "sin", 1.0, 1.0, 0.0, 0.0): ['http:localhost.com']
}


class ChainOfIntegrators:
    """Control-bounded Chain-of-integrators ADC example.

    This class instantiates the chain-of-integrators control-bounded ADC
    which was frequently used in

    * `H. Malmberg, Control-bounded converters, Ph.D. dissertation, Dept. Inf. Technol. Elect. Eng., ETH Zurich, 2020.  <https://www.research-collection.ethz.ch/bitstream/handle/20.500.11850/469192/control-bounded_converters_a_dissertation_by_hampus_malmberg.pdf>`_
    * `H.-A. Loeliger, H. Malmberg, and G. Wilckens, Control-bounded analog-to-digital conversion: transfer function analysis, proof of concept, and digital filter implementation," arXiv:2001.05929 <https://arxiv.org/abs/2001.05929>`_
    * `H.-A. Loeliger and G. Wilckens, Control-based analog-to-digital conversion without sampling and quantization, 2015 Information Theory & Applications Workshop (ITA), UCSD, La Jolla, CA, USA, Feb. 1-6, 2015 <https://ieeexplore.ieee.org/document/7308975>`_
    * `H.-A. Loeliger, L. Bolliger, G. Wilckens, and J. Biveroni, Analog-to-digital conversion using unstable filters, 2011 Information Theory & Applications Workshop (ITA), UCSD, La Jolla, CA, USA, Feb. 6-11, 2011 <https://ieeexplore.ieee.org/abstract/document/5743620>`_

    Furthermore, many pre-computed control-signals can be conveniently
    be accessed through the :py:func:`cbadc.datasets.chain_of_integrators.ChainOfIntegrators.sin`
    and :py:func:`cbadc.datasets.chain_of_integrators.ChainOfIntegrators.ramp` methods.

    Parameters
    ----------
    N: `int`
        number of analog states, (for the chain-of-integrator this also 
        determines M=N), defaults to N=5.
    beta: `float`
        the integration slope or amplification, defaults to beta=6250.
    rho: `float`
        the local feedback term, defaults to rho=0.
    kappa: `float`
        the control gain, defaults to kappa=1
    """

    def __init__(self, N: int = 5, beta: float = 6250.0, rho: float = 0.0, kappa: float = 1.0) -> None:
        self.size = int(1e12)
        self.N = N
        self.beta = beta
        self.rho = rho
        self.kappa = kappa
        A = beta * np.eye(N, k=-1) + rho * np.eye(N)
        B = np.zeros((N, 1))
        B[0, 0] = beta
        CT = np.eye(N)
        Gamma = - kappa * beta * np.eye(N)
        Gamma_tilde = np.eye(N)
        self.M = self.N
        self.T = 1.0 / (2 * beta)
        self.analog_system = cbadc.analog_system.AnalogSystem(
            A, B, CT, Gamma, Gamma_tilde)
        self.digital_control = cbadc.digital_control.DigitalControl(
            self.T, self.N)

    def sin(self, amplitude: float, frequency: float, phase: float = 0.0, offset: float = 0.0):
        """Provide control signals and simulation settings for a sinusoidal input signal.

        Specifically for an input signal

        :math:`u(t) = \mathrm{amplitude} \cdot \sin(2 \pi \mathrm{frequency} t + \mathrm{phase}) + \mathrm{offset}`

        the resulting simulator and possibly pre-computed control-signals are
        retrived.

        Parameters
        ----------
        amplitude: `float`
        frequency: `float`
            specified in [Hz].
        phase: `float`
            defaults to phase=0.
        offset: `float`
            defaults to offset=0.

        Returns
        -------
        control_signal: Generator[np.ndarray, None, None]
            a control signal sequence, possibly retrived over http.
        simulator: :py:class:`cbadc.simulator.StateSpaceSimulator`
            an instantated simulator.
        size: `int`
            the maximum length of the simulation (1G control signal samples)
        """
        input_signal = cbadc.analog_signal.Sinusodial(
            amplitude, frequency, phase, offset)
        simulator = cbadc.simulator.StateSpaceSimulator(
            self.analog_system, self.digital_control, [input_signal])
        params = (self.N, self.beta, self.rho, self.kappa,
                  "sin", amplitude, frequency, phase, offset)
        if params in chain_of_integrators_pre_simulations:
            control_signal = cbadc.utilities.byte_stream_2_control_signal(
                cbadc.utilities.read_byte_stream_from_url(
                    chain_of_integrators_pre_simulations[params],
                    self.M), self.M)
        else:
            logger.warn(
                "No pre-computed simulation found. Iterating will invlove time consuming simulating.")
            control_signal = simulator
        return control_signal, simulator, self.size

    def ramp(self, amplitude: float, frequency: float, phase: float = 0.0, offset: float = 0.0):
        """Provide control signals and simulation settings for a ramp input signal.

        Specifically for an input signal

        :math:`u(t) = \mathrm{amplitude} \cdot \mathrm{ramp}(t / \mathrm{frequency} + \mathrm{phase}) + \mathrm{offset}`

        where :math:`\mathrm{ramp}(\cdot)` is an unit-scale ramp
        function.

        Parameters
        ----------
        amplitude: `float`
        frequency: `float`
            specified in [Hz].
        phase: `float`
            defaults to phase=0.
        offset: `float`
            defaults to offset=0.

        Returns
        -------
        control_signal: Generator[np.ndarray, None, None]
            a control signal sequence, possibly retrived over http.
        simulator: :py:class:`cbadc.simulator.StateSpaceSimulator`
            an instantated simulator.
        size: `int`
            the maximum length of the simulation (1G control signal samples)
        """
        input_signal = cbadc.analog_signal.Ramp(
            amplitude, frequency, phase, offset)
        simulator = cbadc.simulator.StateSpaceSimulator(
            self.analog_system, self.digital_control, [input_signal])
        params = (self.N, self.beta, self.rho, self.kappa,
                  "ramp", amplitude, frequency, phase, offset)
        if params in chain_of_integrators_pre_simulations:
            control_signal = cbadc.utilities.byte_stream_2_control_signal(
                cbadc.utilities.read_byte_stream_from_url(chain_of_integrators_pre_simulations[params], self.M), self.M)
        else:
            logger.warn(
                "No pre-computed simulation found. Iterating will invlove time consuming simulating.")
            control_signal = simulator
        return control_signal, simulator, self.size



"""Digital estimators.

This module provides alternative implementations for the control-bounded A/D
conterter's digital estimator.
"""
from typing import Iterator
import cbadc
import scipy.linalg
import scipy.integrate
import numpy as np
from numpy.linalg import LinAlgError
import time
import logging
import os

logger = logging.getLogger(__name__)


def bruteForceCare(
        A: np.ndarray, B: np.ndarray,
        Q: np.ndarray, R: np.ndarray) -> np.ndarray:
    timelimit = 10 * 60
    start_time = time.time()
    V = np.eye(A.shape[0])
    V_tmp = np.zeros_like(V)
    tau = 1e-5
    RInv = np.linalg.inv(R)

    while not np.allclose(V, V_tmp, rtol=1e-5, atol=1e-8):
        if time.time() - start_time > timelimit:
            raise Exception("Brute Force CARE solver ran out of time")
        V_tmp = V
        try:
            V = V + tau * (
                np.dot(A, V)
                + np.transpose(np.dot(A, V))
                + Q
                - np.dot(V, np.dot(B, np.dot(RInv, np.dot(B.transpose(), V))))
            )
        except FloatingPointError:
            logger.warning("V_frw:\n{}\n".format(V))
            logger.warning("V_frw.dot(V_frw):\n{}".format(np.dot(V, V)))
            raise FloatingPointError
    return V


def care(
        A: np.ndarray, B: np.ndarray,
        Q: np.ndarray, R: np.ndarray) -> np.ndarray:
    """
    This function solves the forward and backward continuous Riccati equation.
    """
    A = np.array(A, dtype=np.double)
    B = np.array(B, dtype=np.double)
    Q = np.array(Q, dtype=np.double)
    R = np.array(R, dtype=np.double)

    V = np.zeros_like(A)

    try:
        V = scipy.linalg.solve_continuous_are(A, B, Q, R)
    except LinAlgError:
        logger.warning(
            """Cholesky Method Failed for computing the CARE of Vf.
            Starting brute force"""
        )
        V = bruteForceCare(A, B, Q, R)
    return V


class DigitalEstimator(Iterator[np.ndarray]):
    """Batch estimator implementation.

    The digital estimator estimates a filtered version
    :math:`\hat{\mathbf{u}}(t)` (shaped by :py:func:`signal_transfer_function`)
    of the input signal :math:`\mathbf{u}(t)` from a sequence of control
    signals :math:`\mathbf{s}[k]`.

    Specifically, the estimates are computed as

    :math:`\overrightarrow{\mathbf{m}}[k] = \mathbf{A}_f \overrightarrow{\mathbf{m}}[k-1] + \mathbf{B}_f \mathbf{s}[k-1]`,

    :math:`\overleftarrow{\mathbf{m}}[k] = \mathbf{A}_b \overrightarrow{\mathbf{m}}[k+1] + \mathbf{B}_b \mathbf{s}[k]`,

    and

    :math:`\hat{\mathbf{u}}(k T) = \mathbf{W}^\mathsf{T}\\left(\overleftarrow{\mathbf{m}}[k] -  \overrightarrow{\mathbf{m}}[k]\\right)`

    where :math:`\mathbf{A}_f, \mathbf{A}_b \in \mathbb{R}^{N \\times N}`,
    :math:`\mathbf{B}_f, \mathbf{B}_b \in \mathbb{R}^{N \\times M}`, and
    :math:`\mathbf{W}^\mathsf{T} \in \mathbb{R}^{L \\times N}` are the
    precomputed filter coefficient based on the choice of
    :py:class:`cbadc.analog_system.AnalogSystem` and
    :py:class:`cbadc.digital_control.DigitalControl`.

    Parameters
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        an analog system (necessary to compute the estimators filter
        coefficients).
    digital_control : :py:class:`cbadc.digital_control.DigitalControl`
        a digital control (necessary to determine the corresponding DAC
        waveform).
    eta2 : `float`
        the :math:`\eta^2` parameter determines the bandwidth of the estimator.
    K1 : `int`
        batch size.
    K2 : `int`, `optional`
        lookahead size, defaults to 0.
    stop_after_number_of_iterations : `int`
        determine a max number of iterations by the iterator, defaults to
        :math:`2^{63}`.
    Ts: `float`, `optional`
        the sampling time, defaults to the time period of the digital control.
    mid_point: `bool`, `optional`
        set samples in between control updates, i.e., :math:`\hat{u}(kT + T/2)`
        , defaults to False.
    downsample: `int`, `optional`
        set a downsampling factor compared to the control signal rate,
        defaults to 1, i.e., no downsampling.


    Attributes
    ----------

    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        analog system as in :py:class:`cbadc.analog_system.AnalogSystem` or
        from derived class.
    digital_control : :py:class:`cbadc.digital_control.DigitalControl`
        digital control as in :py:class:`cbadc.digital_control.DigitalControl`
        or from derived class.
    eta2 : float
        eta2, or equivalently :math:`\eta^2`, sets the bandwidth of the
        estimator.
    control_signal : :py:class:`cbadc.digital_control.DigitalControl`
        a iterator suppling control signals as
        :py:class:`cbadc.digital_control.DigitalControl`.
    number_of_iterations : `int`
        number of iterations until iterator raises :py:class:`StopIteration`.
    downsample: `int`, `optional`
        The downsampling factor compared to the rate of the control signal.
    mid_point: `bool`
        estimated samples shifted in between control updates, i.e.,
        :math:`\hat{u}(kT + T/2)`.
    K1 : `int`
        number of samples per estimate batch.
    K2 : `int`
        number of lookahead samples per computed batch.
    Ts : `float`
        spacing between samples in seconds.
    Af : `array_like`, shape=(N, N), readonly
        The Af matrix
    Ab : `array_like`, shape=(N, N), readonly
        The Ab matrix
    Bf : `array_like`, shape=(N, M), readonly
        The Bf matrix
    Bb : `array_like`, shape=(N, M), readonly
        The Bb matrix
    WT : `array_like`, shape=(L, N), readonly
        The W matrix transposed

    Yields
    ------
    `array_like`, shape=(L,)
        an input estimate sample :math:`\hat{\mathbf{u}}(t)`

    """

    def __init__(self,
                 analog_system: cbadc.analog_system.AnalogSystem,
                 digital_control: cbadc.digital_control.DigitalControl,
                 eta2: float,
                 K1: int,
                 K2: int = 0,
                 stop_after_number_of_iterations: int = (1 << 63),
                 Ts: float = None,
                 mid_point: bool = False,
                 downsample: int = 1):
        # Check inputs
        if (K1 < 1):
            raise BaseException("K1 must be a positive integer.")
        self.K1 = K1
        if (K2 < 0):
            raise BaseException("K2 must be a non negative integer.")
        self.K2 = K2
        self.K3 = K1 + K2
        self.analog_system = analog_system

        if not np.allclose(
                self.analog_system.D,
                np.zeros_like(self.analog_system.D)
        ):
            raise BaseException(
                """Can't compute filter coefficients for system with non-zero
                D matrix. Consider chaining for removing D""")

        self.digital_control = digital_control
        if(eta2 < 0):
            raise BaseException("eta2 must be non negative.")
        if Ts:
            self.Ts = Ts
        else:
            self.Ts = digital_control.T
        self.eta2 = eta2
        self.control_signal = None

        if (downsample != 1):
            raise NotImplementedError(
                "Downsampling currently not implemented for DigitalEstimator")

        self.number_of_iterations = stop_after_number_of_iterations
        self._iteration = 0
        self._estimate_pointer = self.K1

        # For transfer functions
        self.eta2Matrix = np.eye(self.analog_system.CT.shape[0]) * self.eta2

        self._stop_iteration = False

        self.mid_point = mid_point
        # Initialize filters
        self._compute_filter_coefficients(analog_system, digital_control, eta2)
        self._allocate_memory_buffers()

    def set_iterator(self, control_signal_sequence: Iterator[np.ndarray]):
        """Set iterator of control signals

        Parameters
        -----------
        control_signal_sequence : iterator
            a iterator which outputs a sequence of control signals.
        """
        self.control_signal = control_signal_sequence

    def _compute_filter_coefficients(
            self,
            analog_system: cbadc.analog_system.AnalogSystem,
            digital_control: cbadc.digital_control.DigitalControl,
            eta2: float):
        # Compute filter coefficients
        A = np.array(analog_system.A).transpose()
        B = np.array(analog_system.CT).transpose()
        Q = np.dot(np.array(analog_system.B),
                   np.array(analog_system.B).transpose())
        R = eta2 * np.eye(analog_system.N_tilde)
        # Solve care
        Vf = care(A, B, Q, R)
        Vb = care(-A, B, Q, R)
        CCT: np.ndarray = np.dot(np.array(analog_system.CT).transpose(),
                                 np.array(analog_system.CT))
        tempAf: np.ndarray = analog_system.A - np.dot(Vf, CCT) / eta2
        tempAb: np.ndarray = analog_system.A + np.dot(Vb, CCT) / eta2
        self.Af: np.ndarray = np.asarray(scipy.linalg.expm(tempAf * self.Ts))
        self.Ab: np.ndarray = np.asarray(scipy.linalg.expm(-tempAb * self.Ts))
        Gamma = np.array(analog_system.Gamma)
        # Solve IVPs
        self.Bf: np.ndarray = np.zeros(
            (self.analog_system.N, self.analog_system.M))
        self.Bb: np.ndarray = np.zeros(
            (self.analog_system.N, self.analog_system.M))

        atol = 1e-200
        rtol = 1e-10
        max_step = self.Ts / 1000.0
        if (self.mid_point):
            for m in range(self.analog_system.M):
                def _derivative_forward(t, x):
                    return np.dot(tempAf, x) + \
                        np.dot(Gamma, digital_control.impulse_response(m, t))
                solBf = scipy.integrate.solve_ivp(
                    _derivative_forward,
                    (0, self.Ts / 2.0),
                    np.zeros(
                        self.analog_system.N
                    ),
                    atol=atol, rtol=rtol,
                    max_step=max_step).y[:, -1]

                def _derivative_backward(t, x):
                    return - np.dot(tempAb, x) + \
                        np.dot(Gamma, digital_control.impulse_response(m, t))
                solBb = -scipy.integrate.solve_ivp(
                    _derivative_backward,
                    (0, self.Ts / 2.0),
                    np.zeros(
                        self.analog_system.N
                    ),
                    atol=atol, rtol=rtol,
                    max_step=max_step).y[:, -1]
                for n in range(self.analog_system.N):
                    self.Bf[n, m] = solBf[n]
                    self.Bb[n, m] = solBb[n]
            self.Bf = np.dot(
                np.eye(self.analog_system.N) +
                scipy.linalg.expm(tempAf * self.Ts / 2.0),
                self.Bf
            )
            self.Bb = np.dot(
                np.eye(self.analog_system.N) +
                scipy.linalg.expm(tempAb * self.Ts / 2.0),
                self.Bb
            )
        else:
            for m in range(self.analog_system.M):
                def _derivative_forward_2(t, x):
                    return np.dot(tempAf, x) + \
                        np.dot(Gamma, digital_control.impulse_response(m, t))
                solBf = scipy.integrate.solve_ivp(
                    _derivative_forward_2,
                    (0, self.Ts),
                    np.zeros(self.analog_system.N),
                    atol=atol,
                    rtol=rtol, max_step=max_step).y[:, -1]

                def _derivative_backward_2(t, x):
                    return - np.dot(tempAb, x) + \
                        np.dot(Gamma, digital_control.impulse_response(m, t))
                solBb = -scipy.integrate.solve_ivp(
                    _derivative_backward_2,
                    (0, self.Ts),
                    np.zeros(self.analog_system.N),
                    atol=atol,
                    rtol=rtol,
                    max_step=max_step).y[:, -1]
                self.Bf[:, m] = solBf
                self.Bb[:, m] = solBb
        # self.WT = solve(Vf + Vb, analog_system.B).transpose()
        W, _, _, _ = np.linalg.lstsq(Vf + Vb, analog_system.B, rcond=None)
        self.WT = W.transpose()

    def _allocate_memory_buffers(self):
        # Allocate memory buffers
        self._control_signal = np.zeros(
            (self.K3, self.analog_system.M), dtype=np.int8)
        self._estimate = np.zeros(
            (self.K1, self.analog_system.L), dtype=np.double)
        self._control_signal_in_buffer = 0
        self._mean = np.zeros(
            (self.K1 + 1, self.analog_system.N), dtype=np.double)

    def _compute_batch(self):
        temp_forward_mean = np.zeros(self.analog_system.N, dtype=np.double)
        # check if ready to compute buffer
        if (self._control_signal_in_buffer < self.K3):
            raise BaseException("Control signal buffer not full")
        # compute lookahead
        for k1 in range(self.K3 - 1, self.K1 - 1, -1):
            temp = np.dot(
                self.Ab, self._mean[self.K1, :]) + \
                np.dot(self.Bb, self._control_signal[k1, :])
            for n in range(self.analog_system.N):
                self._mean[self.K1, n] = temp[n]
        # compute forward recursion
        for k2 in range(self.K1):
            temp = np.dot(self.Af, self._mean[k2, :]) + \
                np.dot(self.Bf, self._control_signal[k2, :])
            if (k2 < self.K1 - 1):
                for n in range(self.analog_system.N):
                    self._mean[k2 + 1, n] = temp[n]
            else:
                for n in range(self.analog_system.N):
                    temp_forward_mean[n] = temp[n]
        # compute backward recursion and estimate
        for k3 in range(self.K1 - 1, -1, -1):
            temp = np.dot(
                self.Ab, self._mean[k3 + 1, :]) + \
                np.dot(self.Bb, self._control_signal[k3, :])
            temp_estimate = np.dot(self.WT, temp - self._mean[k3, :])
            self._estimate[k3, :] = temp_estimate[:]
            self._mean[k3, :] = temp[:]
        # reset intital means
        for n in range(self.analog_system.N):
            self._mean[0, n] = temp_forward_mean[n]
            self._mean[self.K1, n] = 0
        # rotate buffer to make place for new control signals
        self._control_signal = np.roll(self._control_signal, -self.K1, axis=0)
        self._control_signal_in_buffer -= self.K1

    def _input(self, s: np.ndarray) -> bool:
        if (self._control_signal_in_buffer == (self.K3)):
            raise BaseException(
                """Input buffer full. You must compute batch before adding
                more control signals""")
        for m in range(self.analog_system.M):
            self._control_signal[self._control_signal_in_buffer, :] = \
                np.asarray(2 * s - 1, dtype=np.int8)
        self._control_signal_in_buffer += 1
        return self._control_signal_in_buffer > (self.K3 - 1)

    def __call__(self, control_signal_sequence: Iterator[np.ndarray]):
        return self.set_iterator(control_signal_sequence)

    def __iter__(self):
        return self

    def __next__(self) -> np.ndarray:
        # Check if control signal iterator is set.
        if self.control_signal is None:
            raise BaseException("No iterator set.")
        # Check if the end of prespecified size
        if(self.number_of_iterations < self._iteration):
            raise StopIteration
        self._iteration += 1

        # Check if there are estimates in the estimate buffer
        if(self._estimate_pointer < self.K1):
            temp = np.array(
                self._estimate[self._estimate_pointer, :], dtype=np.double)
            self._estimate_pointer += 1
            return temp
        # Check if stop iteration has been raised in previous batch
        if (self._stop_iteration):
            logger.warning("Warning: StopIteration received by estimator.")
            raise StopIteration
        # Otherwise start receiving control signals
        full = False

        # Fill up batch with new control signals.
        while (not full):
            # next(self.control_signal) calls the control signal
            # iterator and thus recives new control
            # signal samples
            try:
                control_signal_sample = next(self.control_signal)
            except RuntimeError:
                self._stop_iteration = True
                control_signal_sample = np.zeros(
                    (self.analog_system.M), dtype=np.int8)
            full = self._input(control_signal_sample)

        # Compute new batch of K1 estimates
        self._compute_batch()
        # adjust pointer to indicate that estimate buffer
        # is non empty
        self._estimate_pointer -= self.K1

        # recursively call itself to return new estimate
        return self.__next__()

    def noise_transfer_function(self, omega: np.ndarray):
        """Compute the noise transfer function (NTF) at the angular
        frequencies of the omega array.

        Specifically, computes

        :math:`\\text{NTF}( \omega) = \mathbf{G}( \omega)^\mathsf{H} \\left( \mathbf{G}( \omega)\mathbf{G}( \omega)^\mathsf{H} + \eta^2 \mathbf{I}_N \\right)^{-1}`

        for each angular frequency in omega where where
        :math:`\mathbf{G}(\omega)\in\mathbb{R}^{N \\times L}` is the ATF
        matrix of the analog system and :math:`\mathbf{I}_N` represents a
        square identity matrix.

        Parameters
        ----------
        omega: `array_like`, shape=(K,)
            an array_like object containing the angular frequencies for
            evaluation.

        Returns
        -------
        `array_like`, shape=(L, N_tilde, K)
            return NTF evaluated at K different angular frequencies.
        """
        result = np.zeros(
            (self.analog_system.L, self.analog_system.N_tilde, omega.size))
        for index, o in enumerate(omega):
            G = self.analog_system.transfer_function_matrix(np.array([o]))
            G = G.reshape((self.analog_system.N_tilde, self.analog_system.L))
            GH = G.transpose().conjugate()
            GGH = np.dot(G, GH)
            result[:, :, index] = np.abs(
                np.dot(GH, np.linalg.inv(GGH + self.eta2Matrix)))
        return result

    def signal_transfer_function(self, omega: np.ndarray):
        """Compute the signal transfer function (STF) at the angular
        frequencies of the omega array.

        Specifically, computes

        :math:`\\text{STF}( \omega) = \mathbf{G}( \omega)^\mathsf{H} \\left( \mathbf{G}( \omega)\mathbf{G}( \omega)^\mathsf{H} + \eta^2 \mathbf{I}_N \\right)^{-1} \mathbf{G}( \omega)`

        for each angular frequency in omega where where
        :math:`\mathbf{G}(\omega)\in\mathbb{R}^{N \\times L}` is the ATF
        matrix of the analog system and :math:`\mathbf{I}_N` represents a
        square identity matrix.

        Parameters
        ----------
        omega: `array_like`, shape=(K,)
            an array_like object containing the angular frequencies for
            evaluation.

        Returns
        -------
        `array_like`, shape=(L, K)
            return STF evaluated at K different angular frequencies.
        """
        result = np.zeros((self.analog_system.L, omega.size))
        for index, o in enumerate(omega):
            G = self.analog_system.transfer_function_matrix(
                np.array([o])).reshape(
                (self.analog_system.N_tilde, self.analog_system.L)
            )
            GH = G.transpose().conjugate()
            GGH = np.dot(G, GH)
            result[:, index] = np.abs(
                np.dot(GH, np.dot(np.linalg.inv(GGH + self.eta2Matrix), G)))
        return result

    def control_signal_transfer_function(self, omega: np.ndarray):
        """Compute the control signal transfer function at the angular
        frequencies of the omega array.

        Specifically, computes

        :math:`\\begin{pmatrix}\hat{u}_1(\omega) / s_1(\omega) & \\dots & \hat{u}_1(\omega) / s_M(\omega) \\\ \\vdots & \\ddots & \\vdots \\\ \hat{u}_L(\omega) / s_1(\omega) & \\dots & \hat{u}_L(\omega) / s_M(\omega)  \\end{pmatrix}= \mathbf{G}( \omega)^\mathsf{H} \\left( \mathbf{G}( \omega)\mathbf{G}( \omega)^\mathsf{H} + \eta^2 \mathbf{I}_N \\right)^{-1} \\bar{\mathbf{G}}( \omega)`

        for each angular frequency in omega where where
        :math:`\\bar{\mathbf{G}}( \omega)=  \mathbf{C}^\mathsf{T} \\left(\mathbf{A} - i \omega \mathbf{I}_N\\right)^{-1} \mathbf{\\Gamma} \in\mathbb{R}^{N \\times M}`
        is the transfer function from the control signals to the output and :math:`\mathbf{I}_N` represents a
        square identity matrix.

        Parameters
        ----------
        omega: `array_like`, shape=(K,)
            an array_like object containing the angular frequencies for
            evaluation.

        Returns
        -------
        `array_like`, shape=(L, M, K)
            return STF evaluated at K different angular frequencies.
        """
        result = np.zeros(
            (self.analog_system.L, self.analog_system.M, omega.size))
        for index, o in enumerate(omega):
            G = self.analog_system.transfer_function_matrix(np.array([o])).reshape(
                (self.analog_system.N_tilde, self.analog_system.L))
            G_bar = self.analog_system.control_signal_transfer_function_matrix(
                np.array([o])).reshape(
                (self.analog_system.N_tilde, self.analog_system.M)
            )
            GH = G.transpose().conjugate()
            GGH = np.dot(G, GH)
            result[:, index] = np.abs(
                np.dot(GH, np.dot(np.linalg.inv(GGH + self.eta2Matrix), G_bar)))
        return result

    def __str__(self):
        return f"""Digital estimator is parameterized as
        \neta2 = {self.eta2:.2f}, {10 * np.log10(self.eta2):.0f} [dB],
        \nTs = {self.Ts},\nK1 = {self.K1},\nK2 = {self.K2},
        \nand\nnumber_of_iterations = {self.number_of_iterations}
        \nResulting in the filter coefficients\nAf = \n{self.Af},
        \nAb = \n{self.Ab},
        \nBf = \n{self.Bf},
        \nBb = \n{self.Bb},
        \nand WT = \n{self.WT}."""


class ParallelEstimator(DigitalEstimator):
    """Parallelized batch estimator implementation.

    The parallel estimator estimates a filtered version
    :math:`\hat{\mathbf{u}}(t)` (shaped by :py:func:`signal_transfer_function`)
    of the input signal :math:`\mathbf{u}(t)` from a sequence of control
    signals :math:`\mathbf{s}[k]`.

    Specifically, the parallel estimator is a modified version of the default
    estimator :py:class:`cbadc.digital_estimator.DigitalEstimator` where the
    the filter matrices are diagonalized enabling a more efficient and
    possible parallelizable filter implementation. The estimate is computed as

    :math:`\hat{\mathbf{u}}(k T)[\ell] = \sum_{n=0}^N f_w[n] \cdot \overrightarrow{\mathbf{m}}[k][n] + b_w[n] \cdot \overleftarrow{\mathbf{m}}[k][n]`

    where

    :math:`\overrightarrow{\mathbf{m}}[k][n] = f_a[n] \cdot \overrightarrow{\mathbf{m}}[k-1][n] + \sum_{m=0}^{M-1} f_b[n, m] \cdot \mathbf{s}[k-1][m]`

    and

    :math:`\overleftarrow{\mathbf{m}}[k][n] = b_a \cdot \overrightarrow{\mathbf{m}}[k+1][n] + \sum_{m=0}^{M-1} b_b[n, m] \cdot \mathbf{s}[k][m]`.

    Furthermore, :math:`f_a, b_a \in \mathbb{R}^{N}`, :math:`f_b, b_b \in \mathbb{R}^{N \\times M}`,
    and :math:`f_w, b_w \in \mathbb{R}^{L \\times N}` are the precomputed filter coefficient formed
    from the filter coefficients as in :py:class:`cbadc.digital_estimator.DigitalEstimator`.

    Parameters
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        an analog system (necessary to compute the estimators filter coefficients).
    digital_control : :py:class:`cbadc.digital_control.DigitalControl`
        a digital control (necessary to determine the corresponding DAC waveform).
    eta2 : `float`
        the :math:`\eta^2` parameter determines the bandwidth of the estimator.
    K1 : `int`
        batch size.
    K2 : `int`, `optional`
        lookahead size, defaults to 0.
    stop_after_number_of_iterations : `int`
        determine a max number of iterations by the iterator, defaults to :math:`2^{63}`.
    Ts: `float`, `optional`
        the sampling time, defaults to the time period of the digital control.
    mid_point: `bool`, `optional`
        set samples in between control updates, i.e., :math:`\hat{u}(kT + T/2)`, defaults to False.
    downsample: `int`, `optional`
        set a downsampling factor compared to the control signal rate, defaults to 1, i.e.,
        no downsampling.


    Attributes
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        analog system as in :py:class:`cbadc.analog_system.AnalogSystem` or from
        derived class.
    digital_control : :py:class:`cbadc.digital_control.DigitalControl`
        digital control as in :py:class:`cbadc.digital_control.DigitalControl` or from
        derived class.
    eta2 : float
        eta2, or equivalently :math:`\eta^2`, sets the bandwidth of the estimator.
    control_signal : :py:class:`cbadc.digital_control.DigitalControl`
        a iterator suppling control signals as :py:class:`cbadc.digital_control.DigitalControl`.
    number_of_iterations : `int`
        number of iterations until iterator raises :py:class:`StopIteration`.
    downsample: `int`, `optional`
        The downsampling factor compared to the rate of the control signal.
    mid_point: `bool`
        estimated samples shifted in between control updates, i.e., :math:`\hat{u}(kT + T/2)`.
    K1 : `int`
        number of samples per estimate batch.
    K2 : `int`
        number of lookahead samples per computed batch.
    Ts : `float`
        spacing between samples in seconds.
    Af : `array_like`, shape=(N, N), readonly
        The Af matrix
    Ab : `array_like`, shape=(N, N), readonly
        The Ab matrix
    Bf : `array_like`, shape=(N, M), readonly
        The Bf matrix
    Bb : `array_like`, shape=(N, M), readonly
        The Bb matrix
    WT : `array_like`, shape=(L, N), readonly
        The W matrix transposed.
    f_a : `array_like`, shape=(N), readonly
        The :math:`f_a` vector.
    b_a : `array_like`, shape=(N), readonly
        The :math:`b_a` vector.
    f_b : `array_like`, shape=(N, M), readonly
        The :math:`f_b` matrix.
    b_b : `array_like`, shape=(N, M), readonly
        The :math:`b_b` matrix.
    f_w : `array_like`, shape=(L, N), readonly
        The :math:`f_w` matrix.
    b_w : `array_like`, shape=(L, N), readonly
        The :math:`b_w` matrix.

    Yields
    ------
    `array_like`, shape=(L,)
        an input estimate sample :math:`\hat{\mathbf{u}}(t)`
    """

    def __init__(self,
                 analog_system: cbadc.analog_system.AnalogSystem,
                 digital_control: cbadc.digital_control.DigitalControl,
                 eta2: float,
                 K1: int,
                 K2: int = 0,
                 stop_after_number_of_iterations: int = (1 << 63),
                 Ts: float = None,
                 mid_point: bool = False,
                 downsample: int = 1):
        # Check inputs
        if (K1 < 1):
            raise BaseException("K1 must be a positive integer.")
        self.K1 = K1
        if (K2 < 0):
            raise BaseException("K2 must be a non negative integer.")
        self.K2 = K2
        self.K3 = K1 + K2
        self.analog_system = analog_system
        self.digital_control = digital_control
        if(eta2 < 0):
            raise BaseException("eta2 must be non negative.")
        if Ts:
            self.Ts = Ts
        else:
            self.Ts = digital_control.T
        self.eta2 = eta2
        self.control_signal = None

        if (downsample != 1):
            raise NotImplementedError(
                "Downsampling currently not implemented for ParallelEstimator")

        self.number_of_iterations = stop_after_number_of_iterations
        self._iteration = 0
        self._estimate_pointer = self.K1

        # For transfer functions
        self.eta2Matrix = np.eye(self.analog_system.CT.shape[0]) * self.eta2

        self._stop_iteration = False

        self.mid_point = mid_point

        # Initialize filters
        self._compute_filter_coefficients(analog_system, digital_control, eta2)
        self._allocate_memory_buffers()

    def _compute_filter_coefficients(
            self,
            analog_system: cbadc.analog_system.AnalogSystem,
            digital_control: cbadc.digital_control.DigitalControl,
            eta2: float):
        # Compute filter coefficients from base class
        DigitalEstimator._compute_filter_coefficients(
            self, analog_system, digital_control, eta2)
        # Parallelize
        temp, Q_f = np.linalg.eig(self.Af)
        self.forward_a = np.array(temp, dtype=np.complex128)
        Q_f_inv = np.linalg.pinv(Q_f, rcond=1e-20)
        temp, Q_b = np.linalg.eig(self.Ab)
        self.backward_a = np.array(temp, dtype=np.complex128)
        Q_b_inv = np.linalg.pinv(Q_b, rcond=1e-20)

        self.forward_b = np.array(
            np.dot(Q_f_inv, self.Bf), dtype=np.complex128)
        self.backward_b = np.array(
            np.dot(Q_b_inv, self.Bb), dtype=np.complex128)

        self.forward_w = -np.array(np.dot(self.WT, Q_f), dtype=np.complex128)
        self.backward_w = np.array(np.dot(self.WT, Q_b), dtype=np.complex128)

    def _allocate_memory_buffers(self):
        # Allocate memory buffers
        self._control_signal = np.zeros(
            (self.K3, self.analog_system.M), dtype=np.int8)
        self._estimate = np.zeros(
            (self.K1, self.analog_system.L), dtype=np.double)
        self._control_signal_in_buffer = 0
        self._mean = np.zeros((self.analog_system.N), dtype=np.complex128)

    def _compute_batch(self):
        mean: np.complex128 = np.complex128(0)
        # check if ready to compute buffer
        if (self._control_signal_in_buffer < self.K3):
            raise BaseException("Control signal buffer not full")

        self._estimate = np.zeros(
            (self.K1, self.analog_system.L), dtype=np.double)

        for n in range(self.analog_system.N):
            mean = self._mean[n]
            for k1 in range(self.K1):
                for l in range(self.analog_system.L):
                    self._estimate[k1,
                                   l] += np.real(self.forward_w[l, n] * mean)
                mean = self.forward_a[n] * mean
                for m in range(self.analog_system.M):
                    if(self._control_signal[k1, m]):
                        mean += self.forward_b[n, m]
                    else:
                        mean -= self.forward_b[n, m]
            self._mean[n] = mean
            mean = np.complex128(0.0)
            for k3 in range(self.K3 - 1, -1, -1):
                mean = self.backward_a[n] * mean
                for m in range(self.analog_system.M):
                    if(self._control_signal[k3, m]):
                        mean += self.backward_b[n, m]
                    else:
                        mean -= self.backward_b[n, m]
                if (k3 < self.K1):
                    for l in range(self.analog_system.L):
                        self._estimate[k3,
                                       l] += np.real(self.backward_w[l, n] * mean)
        self._control_signal = np.roll(self._control_signal, -self.K1, axis=0)
        self._control_signal_in_buffer -= self.K1

    def _input(self, s: np.ndarray) -> bool:
        if (self._control_signal_in_buffer == (self.K3)):
            raise BaseException(
                "Input buffer full. You must compute batch before adding more control signals")
        self._control_signal[self._control_signal_in_buffer,
                             :] = np.asarray(s, dtype=np.int8)
        self._control_signal_in_buffer += 1
        return self._control_signal_in_buffer > (self.K3 - 1)

    def __iter__(self):
        return self

    def __next__(self) -> np.ndarray:
        # Check if control signal iterator is set.
        if self.control_signal is None:
            raise BaseException("No iterator set.")

        # Check if the end of prespecified size
        if(self.number_of_iterations < self._iteration):
            raise StopIteration
        self._iteration += 1

        # Check if there are estimates in the estimate buffer
        if(self._estimate_pointer < self.K1):
            temp = np.array(
                self._estimate[self._estimate_pointer, :], dtype=np.double)
            self._estimate_pointer += 1
            return temp
        # Check if stop iteration has been raised in previous batch
        if (self._stop_iteration):
            logger.warning("StopIteration received by estimator.")
            raise StopIteration
        # Otherwise start receiving control signals
        full = False

        # Fill up batch with new control signals.
        while (not full):
            # next(self.control_signal) calls the control signal
            # generator and thus recives new control
            # signal samples
            try:
                control_signal_sample = next(self.control_signal)
            except RuntimeError:
                self._stop_iteration = True
                control_signal_sample = np.zeros(
                    (self.analog_system.M), dtype=np.int8)
            full = self._input(control_signal_sample)

        # Compute new batch of K1 estimates
        self._compute_batch()
        # adjust pointer to indicate that estimate buffer
        # is non empty
        self._estimate_pointer -= self.K1

        # recursively call itself to return new estimate
        return self.__next__()

    def __str__(self):
        return f"Parallel estimator is parameterized as \neta2 = {self.eta2:.2f}, {10 * np.log10(self.eta2):.0f} [dB],\nTs = {self.Ts},\nK1 = {self.K1},\nK2 = {self.K2},\nand\nnumber_of_iterations = {self.number_of_iterations}\nResulting in the filter coefficients\nf_a = \n{self.forward_a},\nb_a = \n{self.backward_b},\nf_b = \n{self.forward_b},\nb_b = \n{self.backward_b},\nf_w = \n{self.forward_w},\nand b_w = \n{self.backward_w}."


class IIRFilter(DigitalEstimator):
    """IIR filter implementation of the digital estimator.

    Specifically, the IIR filter estimator estimates a filtered version :math:`\hat{\mathbf{u}}(t)` (shaped by
    :py:func:`signal_transfer_function`) of the
    input signal :math:`\mathbf{u}(t)` from a sequence of control signals :math:`\mathbf{s}[k]`.

    Specifically, the estimate is of the form

    :math:`\hat{\mathbf{u}}(k T) = - \mathbf{W}^{\mathsf{T}} \overrightarrow{\mathbf{m}}_k + \sum_{\ell=0}^{K_2} \mathbf{h}[\ell] \mathbf{s}[k + \ell]`

    where

    :math:`\mathbf{h}[\ell]=\mathbf{W}^{\mathsf{T}} \mathbf{A}_b^\ell \mathbf{B}_b`

    :math:`\overrightarrow{\mathbf{m}}_k = \mathbf{A}_f \mathbf{m}_{k-1} + \mathbf{B}_f \mathbf{s}[k-1]`

    and :math:`\mathbf{W}^{\mathsf{T}}`, :math:`\mathbf{A}_b`,
    :math:`\mathbf{B}_b`, :math:`\mathbf{A}_f`, and :math:`\mathbf{B}_f`
    are computed based on the analog system, the sample period :math:`T_s`, and the
    digital control's DAC waveform as described in
    # page=67/>`_.
    `control-bounded converters <https://www.research-collection.ethz.ch/bitstream/handle/20.500.11850/469192/control-bounded_converters_a_dissertation_by_hampus_malmberg.pdf?sequence=1&isAllowed=y

    Parameters
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        an analog system (necessary to compute the estimators filter coefficients).
    digital_control : :py:class:`cbadc.digital_control.DigitalControl`
        a digital control (necessary to determine the corresponding DAC waveform).
    eta2 : `float`
        the :math:`\eta^2` parameter determines the bandwidth of the estimator.
    K2 : `int`, `optional`
        lookahead size, defaults to 0.
    stop_after_number_of_iterations : `int`
        determine a max number of iterations by the iterator, defaults to  :math:`2^{63}`.
    Ts: `float`, `optional`
        the sampling time, defaults to the time period of the digital control.
    mid_point: `bool`, `optional`
        set samples in between control updates, i.e., :math:`\hat{u}(kT + T/2)`, defaults to False.
    downsample: `int`, `optional`
        specify down sampling rate in relation to the control period :math:`T`, defaults to 1, i.e.,
        no down sampling.

    Attributes
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        analog system as in :py:class:`cbadc.analog_system.AnalogSystem` or from
        derived class.
    digital_control : :py:class:`cbadc.digital_control.DigitalControl`
        digital control as in :py:class:`cbadc.digital_control.DigitalControl` or from
        derived class.
    eta2 : float
        eta2, or equivalently :math:`\eta^2`, sets the bandwidth of the estimator.
    control_signal : :py:class:`cbadc.digital_control.DigitalControl`
        a iterator suppling control signals as :py:class:`cbadc.digital_control.DigitalControl`.
    number_of_iterations : `int`
        number of iterations until iterator raises :py:class:`StopIteration`.
    mid_point: `bool`
        estimated samples shifted in between control updates, i.e., :math:`\hat{u}(kT + T/2)`.
    K2 : `int`
        number of lookahead samples per computed batch.
    Ts : `float`
        spacing between samples in seconds.
    Af : `array_like`, shape=(N, N)
        The Af matrix.
    Ab : `array_like`, shape=(N, N)
        The Ab matrix.
    Bf : `array_like`, shape=(N, M)
        The Bf matrix.
    Bb : `array_like`, shape=(N, M)
        The Bb matrix.
    WT : `array_like`, shape=(L, N)
        The W matrix transposed.
    h : `array_like`, shape=(L, K2, M)
        filter impulse response.
    downsample: `int`
        down sampling rate in relation to the control period :math:`T`.

    Yields
    ------
    `array_like`, shape=(L,)
        an input estimate sample :math:`\hat{\mathbf{u}}(t)`

    """

    def __init__(self,
                 analog_system: cbadc.analog_system.AnalogSystem,
                 digital_control: cbadc.digital_control.DigitalControl,
                 eta2: float,
                 K2: int,
                 stop_after_number_of_iterations: int = (1 << 63),
                 Ts: float = None,
                 mid_point: bool = False,
                 downsample: int = 1
                 ):
        """Initializes filter coefficients
        """
        if (K2 < 0):
            raise BaseException("K2 must be non negative integer.")
        self.K2 = K2
        self._filter_lag = self.K2 - 1
        self.analog_system = analog_system
        if(eta2 < 0):
            raise BaseException("eta2 must be non negative.")
        self.eta2 = eta2
        self.control_signal = None
        self.number_of_iterations = stop_after_number_of_iterations
        self._iteration = 0
        if Ts:
            self.Ts = Ts
        else:
            self.Ts = digital_control.T

        self.downsample = int(downsample)

        self.mid_point = mid_point

        # For transfer functions
        self.eta2Matrix = np.eye(self.analog_system.CT.shape[0]) * self.eta2

        # Compute filter coefficients
        DigitalEstimator._compute_filter_coefficients(
            self, analog_system, digital_control, eta2)

        # Initialize filter
        self.h = np.zeros((self.analog_system.L, self.K2,
                           self.analog_system.M), dtype=np.double)
        # Compute lookback
        temp2 = np.copy(self.Bb)
        for k2 in range(self.K2):
            self.h[:, k2, :] = np.dot(self.WT, temp2)
            temp2 = np.dot(self.Ab, temp2)
        self._control_signal_valued = np.zeros(
            (self.K2, self.analog_system.M), dtype=np.int8)
        self._mean = np.zeros(self.analog_system.N, dtype=np.double)

    def __iter__(self):
        return self

    def __next__(self) -> np.ndarray:
        # Check if control signal iterator is set.
        if self.control_signal is None:
            raise BaseException("No iterator set.")

        # Check if the end of prespecified size
        self._iteration += 1
        if(self.number_of_iterations and self.number_of_iterations < self._iteration):
            raise StopIteration

        # Rotate control_signal vector
        self._control_signal_valued = np.roll(
            self._control_signal_valued, -1, axis=0)

        # insert new control signal
        try:
            temp = self.control_signal.__next__()
        except RuntimeError:
            logger.warning("Estimator received Stop Iteration")
            raise StopIteration

        self._control_signal_valued[-1,
                                    :] = np.asarray(2 * temp - 1, dtype=np.int8)

        # self._control_signal_valued.shape -> (K2, M)
        # self.h.shape -> (L, K2, M)
        result = - np.dot(self.WT, self._mean)
        self._mean = np.dot(self.Af, self._mean) + \
            np.dot(self.Bf, self._control_signal_valued[0, :])
        if (((self._iteration - 1) % self.downsample) == 0):
            return np.tensordot(self.h, self._control_signal_valued, axes=((1, 2), (0, 1))) + result
            # return np.einsum('ijk,jk', self.h, self._control_signal_valued) + result
        return self.__next__()

    def lookahead(self):
        """Return lookahead size :math:`K2`

        Returns
        -------
        int
            lookahead size.
        """
        return self.K2

    def __str__(self):
        return f"IIR estimator is parameterized as \neta2 = {self.eta2:.2f}, {10 * np.log10(self.eta2):.0f} [dB],\nTs = {self.Ts},\nK2 = {self.K2},\nand\nnumber_of_iterations = {self.number_of_iterations}.\nResulting in the filter coefficients\nAf = \n{self.Af},\nBf = \n{self.Bf},WT = \n{self.WT},\n and h = \n{self.h}."

    def filter_lag(self):
        """Return the lag of the filter.

        As the filter computes the estimate as

        ---------
        |   K2  |
        ---------
        ^
        |
        u_hat[k]


        Returns
        -------
        `int`
            The filter lag.

        """
        return self._filter_lag

    def warm_up(self):
        """Warm up filter by population control signals.

        Specifically fills up internal control signal buffer with
        K2 control signals.
        """
        self._filter_lag += self.K2
        for _ in range(self.K2):
            _ = self.__next__()


class FIRFilter(DigitalEstimator):
    """FIR filter implementation of the digital estimator.

    Specifically, the FIR filter estimator estimates a filtered version :math:`\hat{\mathbf{u}}(t)` (shaped by
    :py:func:`signal_transfer_function`) of the
    input signal :math:`\mathbf{u}(t)` from a sequence of control signals :math:`\mathbf{s}[k]`.

    Specifically, the estimate is of the form

    :math:`\hat{\mathbf{u}}(k T) = \hat{\mathbf{u}}_0 + \sum_{\ell=-K_1}^{K_2} \mathbf{h}[\ell] \mathbf{s}[k + \ell]`

    where

    :math:`\mathbf{h}[\ell]=\\begin{cases}\mathbf{W}^{\mathsf{T}} \mathbf{A}_b^\ell \mathbf{B}_b & \mathrm{if} \, \ell \geq 0 \\\  -\mathbf{W}^{\mathsf{T}} \mathbf{A}_f^{-\ell + 1} \mathbf{B}_f & \mathrm{else} \\end{cases}`

    and :math:`\mathbf{W}^{\mathsf{T}}`, :math:`\mathbf{A}_b`,
    :math:`\mathbf{B}_b`, :math:`\mathbf{A}_f`, and :math:`\mathbf{B}_f`
    are computed based on the analog system, the sample period :math:`T_s`, and the
    digital control's DAC waveform as described in
    # page=67/>`_.
    `control-bounded converters <https://www.research-collection.ethz.ch/bitstream/handle/20.500.11850/469192/control-bounded_converters_a_dissertation_by_hampus_malmberg.pdf?sequence=1&isAllowed=y

    Parameters
    ----------
    analog_system : :py:class:`cbadc.analog_system.AnalogSystem`
        an analog system (necessary to compute the estimators filter coefficients).
    digital_control: :py:class:`cbadc.digital_control.DigitalControl`
        a digital control (necessary to determine the corresponding DAC waveform).
    eta2: `float`
        the :math:`\eta^2` parameter determines the bandwidth of the estimator.
    K1: `int`
        The lookback size
    K2: `int`, `optional`
        lookahead size, defaults to 0.
    stop_after_number_of_iterations: `int`
        determine a max number of iterations by the iterator, defaults to  :math:`2^{63}`.
    Ts: `float`, `optional`
        the sampling time, defaults to the time period of the digital control.
    mid_point: `bool`, `optional`
        set samples in between control updates, i.e., :math:`\hat{u}(kT + T/2)`, defaults to False.
    downsample: `int`, `optional`
        specify down sampling rate in relation to the control period :math:`T`, defaults to 1, i.e.,
        no down sampling.
    offset: `array_like`, shape=(L)
        the estimate offset :math:`\hat{\mathbf{u}}_0`, defaults to a zero vector.

    Attributes
    ----------
    analog_system: :py:class:`cbadc.analog_system.AnalogSystem`
        analog system as in :py:class:`cbadc.analog_system.AnalogSystem` or from
        derived class.
    eta2: float
        eta2, or equivalently :math:`\eta^2`, sets the bandwidth of the estimator.
    control_signal: :py:class:`cbadc.digital_control.DigitalControl`
        a iterator suppling control signals as :py:class:`cbadc.digital_control.DigitalControl`.
    number_of_iterations: `int`
        number of iterations until iterator raises :py:class:`StopIteration`.
    K1: `int`
        number of samples, prior to estimate, used in estimate
    K2: `int`
        number of lookahead samples per computed batch.
    Ts: `float`
        spacing between samples in seconds.
    mid_point: `bool`
        estimated samples shifted in between control updates, i.e., :math:`\hat{u}(kT + T/2)`.
    downsample: `int`, `optional`
        down sampling rate in relation to the control period :math:`T`.
    Af: `array_like`, shape=(N, N)
        The Af matrix
    Ab: `array_like`, shape=(N, N)
        The Ab matrix
    Bf: `array_like`, shape=(N, M)
        The Bf matrix
    Bb: `array_like`, shape=(N, M)
        The Bb matrix
    WT: `array_like`, shape=(L, N)
        The W matrix transposed
    h: `array_like`, shape=(L, K1 + K2, M)
        filter impulse response
    offset: `array_like`, shape=(L)
        the estimate offset :math:`\hat{\mathbf{u}}_0`.

    Yields
    ------
    `array_like`, shape=(L,)
        an input estimate sample :math:`\hat{\mathbf{u}}(t)`
    """

    def __init__(self,
                 analog_system: cbadc.analog_system.AnalogSystem,
                 digital_control: cbadc.digital_control.DigitalControl,
                 eta2: float,
                 K1: int,
                 K2: int,
                 stop_after_number_of_iterations: int = (1 << 63),
                 Ts: float = None,
                 mid_point: bool = False,
                 downsample: int = 1,
                 offset: np.ndarray = None):
        """Initializes filter coefficients
        """
        if (K1 < 0):
            raise BaseException("K1 must be non negative integer.")
        self.K1 = K1
        if (K2 < 1):
            raise BaseException("K2 must be a positive integer.")
        self.K2 = K2
        self.K3 = K1 + K2
        self._filter_lag = self.K2 - 1
        self.analog_system = analog_system
        self.digital_control = digital_control
        if(eta2 < 0.0):
            raise BaseException("eta2 must be non negative.")
        self.eta2 = eta2
        self.control_signal = None
        self.number_of_iterations = stop_after_number_of_iterations
        self._iteration = 0
        if Ts:
            self.Ts = Ts
        else:
            self.Ts = digital_control.T
        if(mid_point):
            raise NotImplementedError("Planned for v.0.1.0")
        self.mid_point = mid_point
        self.downsample = int(downsample)
        self._temp_controls = np.zeros(
            (self.downsample, self.analog_system.M), dtype=np.int8)

        if offset is not None:
            self.offset = np.array(offset, dtype=np.float64)
            if (self.offset.size != self.analog_system.L):
                raise BaseException("offset is not of size L")
        else:
            self.offset = np.zeros(self.analog_system.L, dtype=np.float64)

        # For transfer functions
        self.eta2Matrix = np.eye(self.analog_system.CT.shape[0]) * self.eta2
        # Compute filter coefficients
        DigitalEstimator._compute_filter_coefficients(
            self, analog_system, digital_control, eta2)

        # Initialize filter.
        self.h = np.zeros((self.analog_system.L, self.K3,
                           self.analog_system.M), dtype=np.double)
        # Compute lookback.
        temp1 = np.copy(self.Bf)
        for k1 in range(self.K1 - 1, -1, -1):
            self.h[:, k1, :] = - np.dot(self.WT, temp1)
            temp1 = np.dot(self.Af, temp1)
        # Compute lookahead.
        temp2 = np.copy(self.Bb)
        for k2 in range(self.K1, self.K3):
            self.h[:, k2, :] = np.dot(self.WT, temp2)
            temp2 = np.dot(self.Ab, temp2)
        self._control_signal_valued = np.zeros(
            (self.K3, self.analog_system.M), dtype=np.int8)

    def __iter__(self):
        return self

    def __next__(self) -> np.ndarray:
        # Check if control signal iterator is set.
        if self.control_signal is None:
            raise BaseException("No iterator set.")

        # Check if the end of prespecified size
        self._iteration += self.downsample
        if(self.number_of_iterations and self.number_of_iterations < self._iteration):
            raise StopIteration

        # Rotate control_signal vector
        self._control_signal_valued = np.roll(
            self._control_signal_valued, -self.downsample, axis=0)

        # insert new control signal
        try:
            for index in range(self.downsample):
                self._temp_controls[index, :] = 2 * \
                    self.control_signal.__next__() - 1
        except RuntimeError:
            logger.warning("Estimator received Stop Iteration")
            raise StopIteration

        self._control_signal_valued[self.K3 -
                                    self.downsample:, :] = self._temp_controls

        # self._control_signal_valued.shape -> (K3, M)
        # self.h.shape -> (L, K3, M)
        return np.tensordot(self.h, self._control_signal_valued, axes=((1, 2), (0, 1))) + self.offset
        # return np.einsum('lkm,km', self.h, self._control_signal_valued)
        # the Einstein summation results in:
        # result = np.zeros(self._L)
        # for l in range(self._L):
        #    for k in range(self.K1 + self.K2):
        #        for m in range(self._M):
        #            result[l] += self.h[l, k, m] * self._control_signal_valued[k, m]
        # return result

    def lookback(self):
        """Return lookback size :math:`K1`.

        Returns
        -------
        int
            lookback size.
        """
        return self.K1

    def __str__(self):
        return f"FIR estimator is parameterized as \neta2 = {self.eta2:.2f}, {10 * np.log10(self.eta2):.0f} [dB],\nTs = {self.Ts},\nK1 = {self.K1},\nK2 = {self.K2},\nand\nnumber_of_iterations = {self.number_of_iterations}.\nResulting in the filter coefficients\nh = \n{self.h}."

    def filter_lag(self):
        """Return the lag of the filter.

        As the filter computes the estimate as
        -----------------
        |   K1  |   K2  |
        -----------------
                ^
                |
                u_hat[k]


        Returns
        -------
        `int`
            The filter lag.

        """
        return self._filter_lag

    def warm_up(self):
        """Warm up filter by population control signals.

        Specifically fills up internal control signal buffer with
        K2 control signals.
        """
        self._filter_lag += self.K3
        for _ in range(self.K3):
            _ = self.__next__()

    def fir_filter_transfer_function(self, Ts: float = 1.0):
        """Compute the FFT of the system impulse response (FIR filter coefficients).

        Parameters
        ----------
        Ts: `float`
            the sample period of the corresponding impulse response.

        Returns
        -------
        `array_like`, shape=(L, K3 // 2, M)
            the FFT of the corresponding impulse responses.
        """
        frequency_response = np.fft.rfft(self.h, axis=1)
        frequencies = np.fft.rfftfreq(self.K3, d=Ts)
        return (frequencies, frequency_response)

    def convolve(self, filter: np.ndarray):
        """Shape :math:`\mathbf{h}` filter by convolving with filter

        Parameters
        ----------
        filter: `array_like`, shape=(K)
            filter to be applied for each digital control filter
            equivalently.
        """
        for l in range(self.analog_system.L):
            for m in range(self.analog_system.M):
                # self.h.shape -> (L, K3, M)
                temp = np.convolve(
                    self.h[l, :, m], filter, mode='full')
                if temp.size == self.K3:
                    self.h[l, :, m] = temp
                else:
                    mid_point = temp.size // 2
                    half_length = self.K3 // 2
                    self.h[l, :, m] = temp[(mid_point -
                                            half_length):(mid_point + half_length)]

    def write_C_header(self, filename: str):
        """Write the FIR filter coefficients h into
        a C header file.

        Parameters
        ----------
        filename: `str`
            filename of header file.
        """
        if not filename.endswith(".h"):
            filename += ".h"

        with open(filename, 'w') as f:
            f.write(
                "// This file was autogenerated using the src/cbadc/digital_estimator.py FIRFilter class's write_C_header function." + os.linesep)
            f.write(f"#define L {self.analog_system.L}" + os.linesep)
            f.write(f"#define M {self.analog_system.M}" + os.linesep)
            f.write(f"#define K1 {self.K1}" + os.linesep)
            f.write(f"#define K2 {self.K2}" + os.linesep)
            f.write(f"#define K3 {self.K3}" + os.linesep)
            # self.h.shape -> (L, K3, M)
            f.write("double h[L][M][K3] = ")
            f.write("{")
            for l in range(self.analog_system.L):
                f.write("{")
                for m in range(self.analog_system.M):
                    f.write("{")
                    for k in range(self.K3 - 1):
                        f.write(f"{self.h[l,k,m]:.17E},")
                    f.write(f"{self.h[l,-1,m]:.17E}" + "}")
                    if m < (self.analog_system.M - 1):
                        f.write(",")
                f.write("}")
                if l < (self.analog_system.L - 1):
                    f.write(",")
            f.write("}" + os.linesep)


class NUVEstimator():
    """The NUV batch estimator

    The NUV estimator iteratively estimates estimates a filtered version
    :math:`\hat{\mathbf{u}}(t)` of the input signal :math:`\mathbf{u}(t)` from
    a sequence of control signals :math:`\mathbf{s}[k]`.

    In comparision to the linear complexity estimators

    - :py:class:`cbadc.digital_estimator.DigitalEstimator`
    - :py:class:`cbadc.digital_estimator.ParallelEstimator`
    - :py:class:`cbadc.digital_estimator.IIRFilter`
    - :py:class:`cbadc.digital_estimator.FIRFilter`

    the NUV estimator enforces the bounded outputs in an iterative
    scheme as described in

    * `R. Keusch and H.-A. Loeliger, A Binarizing NUV Prior and its Use for M-Level Control and Digital-to-Analog Conversion. <https://arxiv.org/pdf/2105.02599.pdf>`_

    Parameters
    ----------
    analog_system: :py:class:`cbadc.analog_system.AnalogSystem`
        an analog system (necessary to compute the estimators filter
        coefficients).
    digital_control: :py:class:`cbadc.digital_control.DigitalControl`
        a digital control (necessary to determine the corresponding DAC
        waveform).
    covU: `float`
        a prior covariance matrix for the sought estimate.
    bound_y: `float`
        the bound enforced on the outputs :math:`\mathbf{y}(t)` of the
        system.
    gamma: `float`
        a scale factor.
    K1 : `int`
        batch size.
    K2 : `int`, `optional`
        lookahead size, defaults to 0.
    iterations_per_batch: `int`, `optional`
        number of iteration steps per batch, defaults to 100.
    Ts: `float`, `optional`
        the sampling time, defaults to the time period of the digital control.
    oversample: `int`, `optional`
        add observations per control signal, defaults to 1, i.e., no oversampling.

    Attributes
    ----------
    analog_system: :py:class:`cbadc.analog_system.AnalogSystem`
        analog system as in :py:class:`cbadc.analog_system.AnalogSystem` or
        from derived class.
    digital_control: :py:class:`cbadc.digital_control.DigitalControl`
        digital control as in :py:class:`cbadc.digital_control.DigitalControl`
        or from derived class.
    K1: `int`
        number of samples per estimate batch.
    K2: `int`
        number of lookahead samples per computed batch.
    covU: `float`
        the prior covariance matrix (or scalar in the single input case)
        of the estimate :math:`\hat{\mathbf{u}}(t)`.
    bound_y: `float`
        the bound enforced on the outputs :math:`\mathbf{y}(t)` of the
        system.
    Ts: `float`
        the sample rate of the estimates
    iterations_per_batch: `int`
        number of iteration steps per batch.
    gamma: `float`
        a scale factor.
    """

    def __init__(self,
                 analog_system: cbadc.analog_system.AnalogSystem,
                 digital_control: cbadc.digital_control.DigitalControl,
                 covU: float,
                 bound_y: float,
                 gamma: float,
                 K1: int,
                 K2: int = 0,
                 iterations_per_batch: int = 100,
                 Ts: float = None,
                 oversample: int = 1):
        # Check inputs
        if (K1 < 1):
            raise BaseException("K1 must be a positive integer.")
        self.K1 = K1
        if (K2 < 0):
            raise BaseException("K2 must be a non negative integer.")
        self.K2 = K2
        self.K3 = K1 + K2
        self.analog_system = analog_system
        self.covU = covU
        self.bound_y = bound_y
        if not np.allclose(self.analog_system.D, np.zeros_like(
                self.analog_system.D)):
            raise BaseException(
                """Can't compute filter coefficients for system with non-zero
                D matrix. Consider chaining for removing D""")

        self.digital_control = digital_control
        if Ts:
            self.Ts = Ts / oversample
        else:
            self.Ts = digital_control.T / oversample
        self.control_signal = None
        self._stop_iteration = False
        self._estimate_pointer = self.K1

        self.iterations_per_batch = iterations_per_batch
        self.gamma = gamma
        # Initialize filters
        self._compute_filter_coefficients(analog_system, digital_control)
        self._allocate_memory_buffers()

        # Set initial Covariances
        self._forward_CoVariance[0, :, :] = 1e12 * np.eye(self.analog_system.N)
        self._oversample = oversample

    def _allocate_memory_buffers(self):
        # Allocate memory buffers
        self._control_signal = np.zeros(
            (
                self.K3,
                self.analog_system.M
            ),
            dtype=np.int8
        )
        self._estimate = np.zeros(
            (
                self.K1,
                self.analog_system.L
            ),
            dtype=np.double
        )
        self._control_signal_in_buffer = 0
        self._forward_mean = np.zeros(
            (
                self.K3,
                self.analog_system.N
            ),
            dtype=np.double
        )
        self._forward_CoVariance = np.zeros(
            (
                self.K3,
                self.analog_system.N,
                self.analog_system.N
            ),
            dtype=np.double
        )
        self._G = np.zeros(
            (
                self.K3,
                self.analog_system.N_tilde,
                self.analog_system.N_tilde
            ),
            dtype=np.double
        )
        self._F = np.zeros(
            (
                self.K3,
                self.analog_system.N,
                self.analog_system.N
            ),
            dtype=np.double
        )

        self._y_mean = np.zeros(
            (
                self.K3,
                self.analog_system.N_tilde
            )
        )
        self._y_CoVariance = np.zeros(
            (
                self.K3,
                self.analog_system.N_tilde,
                self.analog_system.N_tilde
            )
        )
        # self._W_tilde = np.zeros(
        #     (
        #         self.K3,
        #         self.analog_system.N,
        #         self.analog_system.N
        #     ),
        #     dtype=np.double
        # )
        self._xi_tilde = np.zeros(
            (
                self.K3 + 1,
                self.analog_system.N

            ),
            dtype=np.double
        )
        self._sigma_squared_1 = np.ones(
            (
                self.K3,
                self.analog_system.N_tilde,
            ), dtype=np.double
        )
        self._sigma_squared_2 = np.ones(
            (
                self.K3,
                self.analog_system.N_tilde,
            ), dtype=np.double
        )
        self._posterior_observation_mean = np.zeros(
            (
                self.K3,
                self.analog_system.N_tilde
            ), dtype=np.double
        )

    def set_iterator(self, control_signal_sequence: Iterator[np.ndarray]):
        """Set iterator of control signals

        Parameters
        -----------
        control_signal_sequence : iterator
            a iterator which outputs a sequence of control signals.
        """
        self.control_signal = control_signal_sequence

    def __call__(self, control_signal_sequence: Iterator[np.ndarray]):
        return self.set_iterator(control_signal_sequence)

    def __iter__(self):
        return self

    def __next__(self) -> np.ndarray:
        # Check if control signal iterator is set.
        if self.control_signal is None:
            raise BaseException("No iterator set.")

        # Check if there are estimates in the estimate buffer
        if(self._estimate_pointer < self.K1):
            temp = np.array(
                self._estimate[self._estimate_pointer, :], dtype=np.double)
            self._estimate_pointer += 1
            return temp

        # Check if stop iteration has been raised in previous batch
        if (self._stop_iteration):
            logger.warning("Warning: StopIteration received by estimator.")
            raise StopIteration
        # Otherwise start receiving control signals
        full = False

        # Fill up batch with new control signals.
        while (not full):
            # next(self.control_signal) calls the control signal
            # iterator and thus recives new control
            # signal samples
            try:
                control_signal_sample = next(self.control_signal)
            except RuntimeError:
                self._stop_iteration = True
                control_signal_sample = np.zeros(
                    (self.analog_system.M), dtype=np.int8)
            for _ in range(self._oversample):
                full = self._input(control_signal_sample)

        # Compute new batch of K1 estimates
        self._reset_between_batch()
        self._compute_batch()
        # adjust pointer to indicate that estimate buffer
        # is non empty
        self._estimate_pointer -= self.K1

        # recursively call itself to return new estimate
        return self.__next__()

    def _input(self, s: np.ndarray) -> bool:
        if (self._control_signal_in_buffer == (self.K3)):
            raise BaseException(
                """Input buffer full. You must compute batch before adding
                more control signals""")
        for _ in range(self.analog_system.M):
            self._control_signal[self._control_signal_in_buffer, :] = \
                np.asarray(2 * s - 1, dtype=np.int8)
        self._control_signal_in_buffer += 1
        return self._control_signal_in_buffer > (self.K3 - 1)

    def _compute_filter_coefficients(
            self,
            analog_system: cbadc.analog_system.AnalogSystem,
            digital_control: cbadc.digital_control.DigitalControl,
    ):
        # Compute filter coefficients
        self.Af: np.ndarray = np.asarray(
            scipy.linalg.expm(analog_system.A * self.Ts))
        Gamma = np.array(analog_system.Gamma)
        # Solve IVPs
        self.Bf: np.ndarray = np.zeros(
            (self.analog_system.N, self.analog_system.M))
        atol = 1e-200
        rtol = 1e-10
        max_step = self.Ts / 1000.0

        for m in range(self.analog_system.M):
            def _derivative_forward_2(t, x):
                return np.dot(analog_system.A, x) + \
                    np.dot(Gamma, digital_control.impulse_response(m, t))
            solBf = scipy.integrate.solve_ivp(
                _derivative_forward_2, (0, self.Ts),
                np.zeros(self.analog_system.N), atol=atol,
                rtol=rtol, max_step=max_step
            ).y[:, -1]
            self.Bf[:, m] = solBf

        BBT = np.dot(analog_system.B, analog_system.B.transpose())

        def _derivative_input(t, x):
            t_minus_tau = self.Ts - t
            A_t_minus_tau = analog_system.A * t_minus_tau
            return np.dot(
                scipy.linalg.expm(A_t_minus_tau),
                np.dot(
                    BBT,
                    scipy.linalg.expm(A_t_minus_tau.transpose())
                )
            ).flatten()

        self.Vu = np.dot(
            self.covU,
            scipy.integrate.solve_ivp(
                _derivative_input,
                (0, self.Ts),
                np.zeros(self.analog_system.N ** 2),
                atol=atol,
                rtol=rtol,
                max_step=max_step
            )
            .y[:, -1].reshape(
                (self.analog_system.N, self.analog_system.N)
            )
        )

    def _compute_batch(self):
        # check if ready to compute buffer
        if (self._control_signal_in_buffer < self.K3):
            raise BaseException("Control signal buffer not full")

        self._update_observation_statistics()
        self._MBF()
        for _ in range(self.iterations_per_batch):
            self._update_observation_bound_variances()
            self._update_observation_statistics()
            self._MBF()

        self._input_estimation()

        # Save forward mean and covariance for next batch
        self._forward_mean[0, :] = self._forward_mean[self.K1 - 1, :]
        # Compute posterior mean
        # self._forward_mean[0, :] = self._forward_mean[self.K1 - 1] - np.dot(
        #     self._forward_CoVariance[self.K1 - 1, :, :],
        #     self._xi_tilde[self.K1 - 1, :]
        # )
        self._forward_CoVariance[0, :,
                                 :] = self._forward_CoVariance[self.K1 - 1, :, :]

        # Compute posterior covariance
        # self._forward_CoVariance[0, :, :] = self._forward_CoVariance[self.K1 - 1, :, :] - np.dot(
        #     self._forward_CoVariance[self.K1 - 1, :, :],
        #     np.dot(
        #         self._W_tilde[self.K1 - 1, :, :],
        #         self._forward_CoVariance[self.K1 - 1, :, :]
        #     )
        # )

        self._sigma_squared_1[0, :] = self._sigma_squared_1[self.K1 - 1, :]
        self._sigma_squared_2[0, :] = self._sigma_squared_2[self.K1 - 1, :]

        # rotate buffer to make place for new control signals
        self._control_signal = np.roll(self._control_signal, -self.K1, axis=0)
        self._control_signal_in_buffer -= self.K1

    def _MBF(self):
        # Forward pass
        temp_K3 = self.K3 - 1
        for k in range(self.K3):
            self._G[k, :, :] = np.linalg.inv(
                self._y_CoVariance[k, :, :] + np.dot(
                    self.analog_system.CT,
                    np.dot(
                        self._forward_CoVariance[k, :, :],
                        self.analog_system.CT.transpose()
                    )
                )
            )
            self._F[k, :, :] = np.eye(self.analog_system.N) - np.dot(
                self._forward_CoVariance[k, :, :],
                np.dot(
                    self.analog_system.CT.transpose(),
                    np.dot(
                        self._G[k, :, :],
                        self.analog_system.CT
                    )
                )
            )

            if k < temp_K3:
                self._forward_mean[k + 1, :] = np.dot(
                    self.Af,
                    np.dot(self._F[k, :, :], self._forward_mean[k, :]) +
                    np.dot(
                        self._forward_CoVariance[k, :, :],
                        np.dot(
                            self.analog_system.CT.transpose(),
                            np.dot(
                                self._G[k,
                                        :, :],
                                self._y_mean[k, :]
                            )
                        )
                    )
                ) + \
                    np.dot(
                        self.Bf,
                        self._control_signal[k, :]
                )

                self._forward_CoVariance[k + 1, :, :] = np.dot(
                    self.Af,
                    np.dot(
                        np.dot(
                            self._F[k, :,
                                    :], self._forward_CoVariance[k, :, :]
                        ),
                        self.Af.transpose()
                    )
                ) + self.Vu

        # Backward pass
        for k in range(self.K3 - 1, -1, -1):

            if k < temp_K3:
                xi_tilde_z = np.dot(self.Af.transpose(),
                                    self._xi_tilde[k+1, :])
                # W_tilde_z = np.dot(
                #     self.Af.transpose(),
                #     np.dot(
                #         self._W_tilde[k+1, :, :],
                #         self.Af
                #     )
                # )
            else:
                # Initialize with zero vector.
                xi_tilde_z = np.zeros(self.analog_system.N)
                # W_tilde_z = np.zeros(
                #     (self.analog_system.N, self.analog_system.N))

            # self._W_tilde[k, :, :] = np.dot(
            #     self._F[k, :, :],
            #     np.dot(
            #         W_tilde_z,
            #         self._F[k, :, :]
            #     )
            # ) + np.dot(
            #     self.analog_system.CT.transpose(),
            #     np.dot(
            #         self._G[k, :, :],
            #         self.analog_system.CT
            #     )
            # )

            temp = np.dot(
                self.analog_system.CT.transpose(),
                self._forward_mean[k, :]
            ) - self._y_mean[k, :]

            self._xi_tilde[k, :] = np.dot(
                self._F[k, :, :].transpose(),
                xi_tilde_z
            ) + np.dot(
                self.analog_system.CT.transpose(),
                np.dot(
                    self._G[k, :, :],
                    temp
                )
            )
            # if (k < 5):
            #     print(k,  self._xi_tilde[k, :])

    def _input_estimation(self):
        for k in range(self.K1):
            self._estimate[k, :] = - np.dot(
                self.covU,
                np.dot(
                    self.analog_system.B.transpose(),
                    self._xi_tilde[k, :]
                )
            )

    def _update_observation_bound_variances(self):
        for k in range(self.K3):
            self._posterior_observation_mean[k, :] = np.dot(
                self.analog_system.CT,
                self._forward_mean[k] - np.dot(
                    self._forward_CoVariance[k, :, :],
                    self._xi_tilde[k, :]
                )
            )
            # posterior_CoVariance = np.dot(
            #     self.analog_system.CT,
            #     np.dot(
            #         self._forward_CoVariance[k, :, :] - np.dot(
            #             self._forward_CoVariance[k, :, :],
            #             np.dot(
            #                 self._W_tilde[k, :, :],
            #                 self._forward_CoVariance[k, :, :]
            #             )
            #         ),
            #         self.analog_system.CT.transpose()
            #     )
            # )
            min_sigma_squared = 1e-100
            for ell in range(self.analog_system.N_tilde):
                self._sigma_squared_1[k, ell] = max(
                    np.abs(
                        self._posterior_observation_mean[k, ell] - self.bound_y
                    ) / self.gamma,
                    min_sigma_squared
                )
                self._sigma_squared_2[k, ell] = max(
                    np.abs(
                        self._posterior_observation_mean[k, ell] + self.bound_y
                    ) / self.gamma,
                    min_sigma_squared
                )

    def _update_observation_statistics(self):
        for k in range(self.K3):
            W1_back = np.diag(1.0 / self._sigma_squared_1[k, :])
            W2_back = np.diag(1.0 / self._sigma_squared_2[k, :])

            W_back = W1_back + W2_back
            Wm_back = self.bound_y * \
                np.sum(W1_back, axis=-1) - \
                self.bound_y * np.sum(W2_back, axis=-1)

            self._y_CoVariance[k, :, :] = np.linalg.inv(W_back)

            self._y_mean[k, :] = np.dot(
                self._y_CoVariance[k, :, :],
                Wm_back
            )

    def _reset_between_batch(self):
        self._sigma_squared_1[1:, :] = np.ones(
            (
                self.K3 - 1,
                self.analog_system.N_tilde
            ), dtype=np.double
        )
        self._sigma_squared_2[1:, :] = np.ones(
            (
                self.K3 - 1,
                self.analog_system.N_tilde
            ), dtype=np.double
        )

    def __str__(self):
        return f"""NUVEstimator is parameterized as
        \nobservation bound = {self.bound_y:.2f},
        \ncovU = {self.covU},
        \ngamma = {self.gamma},
        \nTs = {self.Ts},\nK1 = {self.K1},\nK2 = {self.K2},
        \nand\nnumber of iterations per batch = {self.iterations_per_batch}
        \nResulting in the filter coefficients\nAf = \n{self.Af},
        \nBf = \n{self.Bf},
        \nand Vu = \n{self.Vu}.
        """



"""Digital controls

This module provides a general :py:class:`cbadc.digital_control.DigitalControl`
class to enabled the creation of a general independently controlled digital
control system.
"""
import numpy as np


class DigitalControl:
    """Represents a digital control system.

    This is the simplest digital control where
    :math:`M=\\tilde{M}` and each control signal is updated
    independently. Furthermore, the DAC waveform is a constant signal
    as :math:`\mathbf{s}(t)=\mathbf{s}[k]` for :math:`t\in[k T, (k+1)T)`.

    Parameters
    ----------
    T : `float`
        clock period at which the digital control updates.
    M : `int`
        number of controls.
    t0 : `float`: optional
        determines initial time, defaults to 0.

    Attributes
    ----------
    T : `float`
        clock period :math:`T` of digital control system.
    M : `int`
        number of controls :math:`M`.
    M_tilde : `int`
        number of control observations :math:`\\tilde{M}`.

    Note
    ----
    For this digital control system :math:`M=\\tilde{M}`.

    See also
    ---------
    :py:class:`cbadc.simulator.StateSpaceSimulator`

    Examples
    --------
    >>> from cbadc.digital_control import DigitalControl
    >>> T = 1e-6
    >>> M = 4
    >>> dc = DigitalControl(T, M)
    >>> print(dc)
    The Digital Control is parameterized as:
    T = 1e-06,
    M = 4, and next update at
    t = 1e-06
    """

    def __init__(self, T: float, M: int, t0: float = 0.0):
        self.T = T
        self.M = M
        self.M_tilde = M
        self._t_next = t0 + self.T
        self._s = np.zeros(self.M, dtype=np.int8)
        self._dac_values = np.zeros(self.M, dtype=np.double)

    def control_contribution(self, t: float, s_tilde: np.ndarray) -> np.ndarray:
        """Evaluates the control contribution at time t given a control observation
        s_tilde.

        Parameters
        ----------
        t : `float`
            time at which the digital control i evaluated.
        s_tilde : `array_like`, shape=(M_tilde,)
            state vector evaluated at time t

        Examples
        --------
        >>> from cbadc.digital_control import DigitalControl
        >>> import numpy as np
        >>> T = 1e-6
        >>> M = 4
        >>> dc = DigitalControl(T, M)
        >>> res = dc.control_contribution(T + 1e-100, np.array([0.1, -0.2, 0.3, -99]))
        >>> print(np.array(res))
        [ 1. -1.  1. -1.]



        Returns
        -------
        `array_like`, shape=(M,)
            the control signal :math:`\mathbf{s}(t)`

        """
        # Check if time t has passed the next control update
        if t >= self._t_next:
            # if so update the control signal state
            self._s = s_tilde >= 0
            self._t_next += self.T
            # DAC
            self._dac_values = np.asarray(2 * self._s - 1, dtype=np.double)
        return self._dac_values

    def control_signal(self) -> np.ndarray:
        """Returns the current control state, i.e, :math:`\mathbf{s}[k]`.

        Examples
        --------
        >>> from cbadc.digital_control import DigitalControl
        >>> import numpy as np
        >>> T = 1e-6
        >>> M = 4
        >>> dc = DigitalControl(T, M)
        >>> _ = dc.control_contribution(T, np.array([-0.1, -0.2, 0.3, 99]))
        >>> res = dc.control_signal()
        >>> print(np.array(res))
        [0 0 1 1]


        Returns
        -------
        `array_like`, shape=(M,), dtype=numpy.int8
            current control state.
        """
        return self._s

    def impulse_response(self, m: int, t: float):
        """The impulse response of the corresponding DAC waveform

        Parameters
        ----------
        m : `int`
            determines which :math:`m\in\{0,\dots,M-1\}` control dimension
            which is triggered.
        t : `float`
            evaluate the impulse response at time t.

        Returns
        -------
        `array_like`, shape=(M,)
            the dac waveform of the digital control system.

        """
        temp = np.zeros(self.M, dtype=np.double)
        temp[m] = 1
        return temp

    def __str__(self):
        return f"The Digital Control is parameterized as:\nT = {self.T},\nM = {self.M}, and next update at\nt = {self._t_next}"


class PhaseDelayedControl(DigitalControl):
    """Represents a digital control system that switches controls individually
    sequentially.

    This digital control updates the :math:`m`-th control signals as

    :math:`s_m[k] = \\tilde{s}((k+m)T)`

    except for this it works similarly to 
    :py:class`cbadc.digital_control.DigitalControl`

    Parameters
    ----------
    T : `float`
        clock period at which the digital control updates.
    M : `int`
        number of controls.
    t0 : `float`: optional
        determines initial time, defaults to 0.

    Attributes
    ----------
    T : `float`
        clock period :math:`T` of digital control system.
    M : `int`
        number of controls :math:`M`.
    M_tilde : `int`
        number of control observations :math:`\\tilde{M}`.

    Note
    ----
    For this digital control system :math:`M=\\tilde{M}`.
    """

    def __init__(self, T: float, M: int, t0: float = 0):
        DigitalControl.__init__(self, T, M, t0=t0)
        self._t_next = t0 + self.T * np.arange(M)

    def control_contribution(self, t: float, s_tilde: np.ndarray) -> np.ndarray:
        """Evaluates the control contribution at time t given a control observation
        s_tilde.

        Parameters
        ----------
        t : `float`
            time at which the digital control i evaluated.
        s_tilde : `array_like`, shape=(M_tilde,)
            state vector evaluated at time t

        Returns
        -------
        `array_like`, shape=(M,)
            the control signal :math:`\mathbf{s}(t)`

        """
        # Check if time t has passed the next control update
        for m in range(self.M):
            if t >= self._t_next[m]:
                # if so update the control signal state
                self._s[m] = s_tilde[m] >= 0
                self._t_next[m] += self.T * self.M
                # DAC
                self._dac_values = np.asarray(2 * self._s - 1, dtype=np.double)
        return self._dac_values

    def control_signal(self) -> np.ndarray:
        """Returns the current control state, i.e, :math:`\mathbf{s}[k]`.

        Returns
        -------
        `array_like`, shape=(M,), dtype=numpy.int8
            current control state.
        """
        return self._s


class CalibrationControl(DigitalControl):

    def control_contribution(self, t: float, s_tilde: np.ndarray) -> np.ndarray:
        """Evaluates the control contribution at time t given a control observation
        s_tilde.

        Parameters
        ----------
        t : `float`
            time at which the digital control i evaluated.
        s_tilde : `array_like`, shape=(M_tilde,)
            state vector evaluated at time t

        Returns
        -------
        `array_like`, shape=(M,)
            the control signal :math:`\mathbf{s}(t)`

        """
        # Check if time t has passed the next control update
        if t >= self._t_next:
            # if so update the control signal state
            self._s[1:] = s_tilde[1:] >= 0
            # randomize first bit
            self._s[0] = np.random.randint(2)
            self._t_next += self.T
            # DAC
            self._dac_values = np.asarray(
                2 * self._s - 1, dtype=np.double)
        return self._dac_values
